from astroquery.jplhorizons import Horizons

from tom_catalogs.harvester import AbstractHarvester


class JPLHorizonsHarvester(AbstractHarvester):
    """
    The ``JPLHorizonsHarvester`` is the interface to the JPL Horizons catalog. For information regarding the JPL
    Horizons catalog, please see https://ssd.jpl.nasa.gov/?horizons or
    https://astroquery.readthedocs.io/en/latest/jplhorizons/jplhorizons.html.
    """

    name = 'JPL Horizons'

    def query(self, term, location=None, start=None, end=None, step=None):
        if all((start, end, step)):
            epochs = {'start': start, 'end': end, 'step': step}
        else:
            epochs = None
        try:
            obj = Horizons(id=term, location=location, epochs=epochs)
            self.catalog_data = obj.elements()
        except (ValueError, IOError):
            self.catalog_data = {}

    def to_target(self):
        target = super().to_target()
        target.type = 'NON_SIDEREAL'
        target.scheme = 'MPC_MINOR_PLANET'
        target.name = str(self.catalog_data['targetname'][0])
        target.mean_anomaly = self.catalog_data['M'][0]  # mean anomaly in JPL astroquery column names
        target.arg_of_perihelion = self.catalog_data['w'][0]  # argument of the perifocus in JPL
        target.lng_asc_node = self.catalog_data['Omega'][0]  # longitude of asc. node in JPL
        target.inclination = self.catalog_data['incl'][0]  # inclination in JPL
        target.mean_daily_motion = self.catalog_data['n'][0]  # mean motion in JPL
        target.semimajor_axis = self.catalog_data['a'][0]  # semi-major axis in JPL
        target.eccentricity = self.catalog_data['e'][0]  # eccentricity in JPL
        # epoch Julian Date in JPL
        target.epoch_of_elements = self.jd_to_mjd(self.catalog_data['datetime_jd'][0])
        target.epoch_of_perihelion = self.jd_to_mjd(self.catalog_data['Tp_jd'][0])  # time of periapsis in JPL
        target.perihdist = self.catalog_data['q'][0]  # periapsis distance in JPL
        # undocumented in JPL astroquery column names -- presuming P is the orbital period in JPL
        target.ephemeris_period = self.catalog_data['P'][0]
        return target



from django.core.management.base import BaseCommand

from tom_alerts.models import BrokerQuery
from tom_alerts.alerts import get_service_class
from time import sleep


class Command(BaseCommand):
    help = 'Runs saved alert queries and saves the results as Targets'

    def add_arguments(self, parser):
        parser.add_argument(
            'query_name',
            help='One or more saved queries to run'
        )

    def handle(self, *args, **options):
        try:
            query = BrokerQuery.objects.get(name=options['query_name'])
            broker_class = get_service_class(query.broker)
            broker = broker_class()
            alerts = broker.fetch_alerts(query.parameters)
            while True:
                try:
                    generic_alert = broker.to_generic_alert(next(alerts))
                    target = generic_alert.to_target()
                    target.save()
                    self.stdout.write('Created target: {}'.format(target))
                except StopIteration:
                    self.stdout.write('Finished creating targets')
                sleep(1)
        except KeyboardInterrupt:
            self.stdout.write('Exiting...')



from django.conf import settings
from django.contrib import messages
from django.http import HttpResponseForbidden
from django.shortcuts import redirect
from django.urls import reverse

from tom_common.exceptions import ImproperCredentialsException


class ExternalServiceMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        return response

    def process_exception(self, request, exception):
        if isinstance(exception, ImproperCredentialsException):
            msg = (
                    'There was a problem authenticating with {}. Please check that you have the correct '
                    'credentials in the corresponding settings variable. '
                    'https://tom-toolkit.readthedocs.io/en/stable/common/customsettings.html '
                ).format(
                str(exception)
            )
            messages.error(request, msg)
            return redirect(reverse('home'))
        raise exception


class AuthStrategyMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        self.open_urls = [reverse('login')] + getattr(settings, 'OPEN_URLS', [])

    def __call__(self, request):
        if settings.AUTH_STRATEGY == 'LOCKED':
            if not request.user.is_authenticated and request.path_info not in self.open_urls:
                return HttpResponseForbidden()

        return self.get_response(request)


class Raise403Middleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)

        if response.status_code == 403:
            msg = (
                'You do not have permission to access this page. Please login as a user '
                'with the correct permissions or contact your PI.'
            )
            messages.error(request, msg)
            return redirect(reverse('login') + '?next=' + request.path)

        return response



import logging

from django.core.management.base import BaseCommand

from tom_observations.cadence import get_cadence_strategy
from tom_observations.models import DynamicCadence


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    This management command ensures that all cadences are kept up to date. It is intended to be run
    by a cron job, and the frequency should be whatever is determined to be the desired frequency
    by the PI.
    """

    help = 'Entry point for running cadence strategies.'

    def handle(self, *args, **kwargs):
        cadenced_groups = DynamicCadence.objects.filter(active=True)

        updated_cadences = []

        for cg in cadenced_groups:
            try:
                strategy = get_cadence_strategy(cg.cadence_strategy)(cg)
                new_observations = strategy.run()
                if not new_observations:
                    logger.log(msg=f'No changes from dynamic cadence {cg}', level=logging.INFO)
                else:
                    logger.log(msg=f'''Cadence update completed for dynamic cadence {cg},
                                       {len(new_observations)} new observations created.''',
                               level=logging.INFO)
                    updated_cadences.append(cg.observation_group)
            except Exception as e:
                logger.error(msg=f'Unable to run strategy {cg} with id {cg.id} due to error: {e}')

        if updated_cadences:
            msg = 'Created new observations for dynamic cadences with observation groups: {0}.'
            return msg.format(', '.join([str(cg) for cg in updated_cadences]))
        else:
            return 'No new observations for any dynamic cadences.'



from astropy.units import photon, Quantity, spectral_density
from django.db import migrations
from specutils import Spectrum1D

from tom_dataproducts.processors.data_serializers import SpectrumSerializer

"""
This code was written by Curtis McCully for the conversion from photon_flux to flux. The below code is an adaptation
of this code.

def photon_spectrum_to_energy_spectrum(wavelength, photon_counts):
    photon_spectrum = specutils.Spectrum1D(flux=photon_counts, spectral_axis=wavelength)
    energy_spectrum = photon_spectrum.flux * (photon_spectrum.energy / u.photon)
    return specutils.Spectrum1D(spectral_axis=wavelength,
                                flux=energy_spectrum.to('erg / (s cm2 AA)', u.spectral_density(wavelength)))
"""

def photon_spectrum_to_energy_spectrum(apps, schema_editor):
    reduced_datum = apps.get_model('tom_dataproducts', 'ReducedDatum')
    spectrum_serializer = SpectrumSerializer()
    for row in reduced_datum.objects.filter(data_type='spectroscopy'):
        # In order to avoid a KeyError on already-corrected data or data that has no need to be corrected, we only
        # perform the spectroscopy correction on values that have photon_flux/photon_flux_units
        if all(k in row.value.keys() for k in ['photon_flux', 'photon_flux_units', 'wavelength', 'wavelength_units']):
            photon_counts = Quantity(value=row.value['photon_flux'], unit=row.value['photon_flux_units'])
            wavelength = Quantity(value=row.value['wavelength'], unit=row.value['wavelength_units'])
            photon_spectrum = Spectrum1D(flux=photon_counts, spectral_axis=wavelength)
            energy_spectrum = photon_spectrum.flux * (photon_spectrum.energy / photon)
            energy_spectrum_object = Spectrum1D(
                                        spectral_axis=wavelength,
                                        flux=energy_spectrum.to('erg / (s cm2 AA)', spectral_density(wavelength)))
            row.value = spectrum_serializer.serialize(energy_spectrum_object)
            row.save()


class Migration(migrations.Migration):
    dependencies = [
        ('tom_dataproducts', '0009_auto_20210204_2221')
    ]

    operations = [
        migrations.RunPython(photon_spectrum_to_energy_spectrum, reverse_code=migrations.RunPython.noop),
    ]


from django.views.generic.edit import FormView
from django.urls import reverse
from urllib.parse import urlencode
from django.core.exceptions import ValidationError

from .forms import CatalogQueryForm
from .harvester import MissingDataException


class CatalogQueryView(FormView):
    """
    View for querying a specific catalog.
    """

    form_class = CatalogQueryForm
    template_name = 'tom_catalogs/query_form.html'

    def form_valid(self, form):
        """
        Ensures that the form parameters are valid and runs the catalog query.

        :param form: CatalogQueryForm with required parameter`s
        :type form: CatalogQueryForm
        """
        try:
            self.target = form.get_target()
        except MissingDataException:
            form.add_error('term', ValidationError('Object not found'))
            return self.form_invalid(form)
        return super().form_valid(form)

    def get_success_url(self):
        """
        Redirects to the ``TargetCreateView``. Appends the target parameters to the URL as query parameters in order to
        autofill the ``TargetCreateForm``, including any additional names returned from the query.
        """
        target_params = self.target.as_dict()
        target_params['names'] = ','.join(getattr(self.target, 'extra_names', []))
        return reverse('targets:create') + '?' + urlencode(target_params)



from django_filters import rest_framework as drf_filters
from guardian.mixins import PermissionListMixin
from guardian.shortcuts import get_objects_for_user
from rest_framework.mixins import DestroyModelMixin, RetrieveModelMixin
from rest_framework.viewsets import GenericViewSet, ModelViewSet

from tom_targets.filters import TargetFilter
from tom_targets.models import TargetExtra, TargetName
from tom_targets.serializers import TargetSerializer, TargetExtraSerializer, TargetNameSerializer


permissions_map = {  # TODO: Use the built-in DRF mapping or just switch to DRF entirely.
        'GET': 'view_target',
        'OPTIONS': [],
        'HEAD': [],
        'POST': 'add_target',
        'PATCH': 'change_target',
        'PUT': 'change_target',
        'DELETE': 'delete_target'
    }


# Though DRF supports using django-guardian as a permission backend without explicitly using PermissionListMixin, we
# chose to use it because it removes the requirement that a user be granted both object- and model-level permissions,
# and a user that has object-level permissions is understood to also have model-level permissions.
# For whatever reason, get_queryset has to be explicitly defined, and can't be set as a property, else the API won't
# respect permissions.
#
# At present, create is not restricted at all. This seems to be a limitation of django-guardian and should be revisited.
class TargetViewSet(ModelViewSet, PermissionListMixin):
    """
    Viewset for Target objects. By default supports CRUD operations.
    See the docs on viewsets: https://www.django-rest-framework.org/api-guide/viewsets/

    To view supported query parameters, please use the ``OPTIONS`` endpoint, which can be accessed through the web UI.

    **Please note that ``groups`` are an accepted query parameters for the ``CREATE`` endpoint. The ``groups`` parameter
    will specify which ``groups`` can view the created Target. If no ``groups`` are specified, the ``Target`` will only
    be visible to the user that created the ``Target``. Make sure to check your ``groups``!!**

    In order to create new ``TargetName`` or ``TargetExtra`` objects, a dictionary with the new values must be appended
    to the ``aliases`` or ``targetextra_set`` lists. If ``id`` is included, the API will attempt to update an existing
    ``TargetName`` or ``TargetExtra``. If no ``id`` is provided, the API will attempt to create new entries.

    ``TargetName`` and ``TargetExtra`` objects can only be deleted or specifically retrieved via the
    ``/api/targetname/`` or ``/api/targetextra/`` endpoints.
    """
    serializer_class = TargetSerializer
    filter_backends = (drf_filters.DjangoFilterBackend,)
    filterset_class = TargetFilter

    def get_queryset(self):
        permission_required = permissions_map.get(self.request.method)
        return get_objects_for_user(self.request.user, f'tom_targets.{permission_required}')


class TargetNameViewSet(DestroyModelMixin, PermissionListMixin, RetrieveModelMixin, GenericViewSet):
    """
    Viewset for TargetName objects. Only ``GET`` and ``DELETE`` operations are permitted.

    To view available query parameters, please use the OPTIONS endpoint, which can be accessed through the web UI.
    """
    serializer_class = TargetNameSerializer

    def get_queryset(self):
        permission_required = permissions_map.get(self.request.method)
        return TargetName.objects.filter(
            target__in=get_objects_for_user(self.request.user, f'tom_targets.{permission_required}')
        )


class TargetExtraViewSet(DestroyModelMixin, PermissionListMixin, RetrieveModelMixin, GenericViewSet):
    """
    Viewset for TargetExtra objects. Only ``GET`` and ``DELETE`` operations are permitted.

    To view available query parameters, please use the OPTIONS endpoint, which can be accessed through the web UI.
    """
    serializer_class = TargetExtraSerializer

    def get_queryset(self):
        permission_required = permissions_map.get(self.request.method)
        return TargetExtra.objects.filter(
            target__in=get_objects_for_user(self.request.user, f'tom_targets.{permission_required}')
        )



import requests
from django.conf import settings
from django.core.cache import cache

from tom_observations.facilities.lco import LCOFacility, LCOBaseObservationForm
from tom_observations.facilities.lco import LCOImagingObservationForm, LCOSpectroscopyObservationForm
from tom_common.exceptions import ImproperCredentialsException


# Determine settings for this module.
try:
    LCO_SETTINGS = settings.FACILITIES['LCO']
except (AttributeError, KeyError):
    LCO_SETTINGS = {
        'portal_url': 'https://observe.lco.global',
        'api_key': '',
    }

# Module specific settings.
PORTAL_URL = LCO_SETTINGS['portal_url']
TERMINAL_OBSERVING_STATES = ['COMPLETED', 'CANCELED', 'WINDOW_EXPIRED']

# There is currently only one available grating, which is required for spectroscopy.
SPECTRAL_GRATING = 'SYZY_400'


def make_request(*args, **kwargs):
    response = requests.request(*args, **kwargs)
    if 400 <= response.status_code < 500:
        raise ImproperCredentialsException('SOAR: ' + str(response.content))
    response.raise_for_status()
    return response


class SOARBaseObservationForm(LCOBaseObservationForm):

    @staticmethod
    def _get_instruments():
        cached_instruments = cache.get('soar_instruments')

        if not cached_instruments:
            response = make_request(
                'GET',
                PORTAL_URL + '/api/instruments/',
                headers={'Authorization': 'Token {0}'.format(LCO_SETTINGS['api_key'])}
            )

            cached_instruments = {k: v for k, v in response.json().items() if 'SOAR' in k}
            cache.set('soar_instruments', cached_instruments)

        return cached_instruments

    @staticmethod
    def instrument_to_type(instrument_type):
        if 'IMAGER' in instrument_type:
            return 'EXPOSE'
        else:
            return 'SPECTRUM'


class SOARImagingObservationForm(SOARBaseObservationForm, LCOImagingObservationForm):

    @staticmethod
    def instrument_choices():
        return sorted(
            [(k, v['name']) for k, v in SOARImagingObservationForm._get_instruments().items() if 'IMAGE' in v['type']],
            key=lambda inst: inst[1]
        )

    @staticmethod
    def filter_choices():
        return sorted(set([
            (f['code'], f['name']) for ins in SOARImagingObservationForm._get_instruments().values() for f in
            ins['optical_elements'].get('filters', [])
            ]), key=lambda filter_tuple: filter_tuple[1])


class SOARSpectroscopyObservationForm(SOARBaseObservationForm, LCOSpectroscopyObservationForm):

    @staticmethod
    def instrument_choices():
        return sorted(
            [(k, v['name'])
             for k, v in SOARSpectroscopyObservationForm._get_instruments().items()
             if 'SPECTRA' in v['type']],
            key=lambda inst: inst[1])

    @staticmethod
    def filter_choices():
        return set([
            (f['code'], f['name']) for ins in SOARSpectroscopyObservationForm._get_instruments().values() for f in
            ins['optical_elements'].get('slits', [])
            ])

    def _build_instrument_config(self):
        instrument_configs = super()._build_instrument_config()

        instrument_configs[0]['optical_elements'] = {
            'slit': self.cleaned_data['filter'],
            'grating': SPECTRAL_GRATING
        }
        instrument_configs[0]['rotator_mode'] = 'SKY'

        return instrument_configs


class SOARFacility(LCOFacility):
    """
    The ``SOARFacility`` is the interface to the SOAR Telescope. For information regarding SOAR observing and the
    available parameters, please see http://www.ctio.noao.edu/soar/content/observing-soar.

    Please note that SOAR is only available in AEON-mode. It also uses the LCO API key, so to use this module, the
    LCO dictionary in FACILITIES in `settings.py` will need to be completed.
    """

    name = 'SOAR'
    observation_forms = {
        'IMAGING': SOARImagingObservationForm,
        'SPECTRA': SOARSpectroscopyObservationForm
    }
    # The SITES dictionary is used to calculate visibility intervals in the
    # planning tool. All entries should contain latitude, longitude, elevation
    # and a code.
    SITES = {
        'Cerro Pachón': {
            'sitecode': 'sor',
            'latitude': -30.237892,
            'longitude': -70.733642,
            'elevation': 2000
        }
    }

    def get_form(self, observation_type):
        return self.observation_forms.get(observation_type, SOARBaseObservationForm)



import logging

from django.conf import settings

from tom_observations.facility import BaseManualObservationFacility, BaseManualObservationForm
from tom_targets.models import Target

logger = logging.getLogger(__name__)


#
# facility properties needed by both the Facility and Form classes
# are candidates for module-level definitions. If the property is just
# for the Facility, put it in the class definition
#

try:
    EXAMPLE_MANUAL_SETTINGS = settings.FACILITIES['EXAMPLE_MANUAL']
except KeyError:
    EXAMPLE_MANUAL_SETTINGS = {
    }

EXAMPLE_SITES = {
    'Example Manual Facility': {
        'sitecode': 'Example',
        'latitude': 0.0,
        'longitude': 0.0,
        'elevation': 0.0
    },
}
EXAMPLE_TERMINAL_OBSERVING_STATES = ['Completed']


class ExampleManualFacility(BaseManualObservationFacility):
    """
    """

    name = 'Example'
    observation_types = [('OBSERVATION', 'Manual Observation')]

    def get_form(self, observation_type):
        """
        This method takes in an observation type and returns the form type that matches it.
        """
        return BaseManualObservationForm

    def submit_observation(self, observation_payload):
        """
        This method takes in the serialized data from the form.

        The BaseManualObservationForm(BaseObservationForm) does not require an observation_id.
        In this example, if no observation_id is given, we construct one to return from the
        other required form fields.

        """
        # TODO: explore adding logic to send email to tom-demo

        obs_ids = []
        # params comes as JSON string, to turn it back into a dictionary
        obs_params = observation_payload['params']

        # if the Observation id was supplied then use it
        if obs_params['observation_id']:
            obs_ids.append(obs_params['observation_id'])
        else:
            # observation_id was empty string, so construct reasonable default
            # such as name:target-facility-start
            target = Target.objects.get(pk=observation_payload['target_id']).name
            obs_name = obs_params['name']
            facility = obs_params['facility']
            start = obs_params[self.get_start_end_keywords()[0]]

            obs_id = f'{obs_name}:{target}-{facility}-{start}'
            obs_ids.append(obs_id)

        return obs_ids

    def validate_observation(self, observation_payload):
        """
        Same thing as submit_observation, but a dry run. You can
        skip this in different modules by just using "pass"
        """
        raise NotImplementedError

    def is_fits_facility(self, header):
        """
        Returns True if the FITS header is from this facility based on valid keywords and associated
        values, False otherwise.
        """
        return False

    def get_start_end_keywords(self):
        """
        Returns the keywords representing the start and end of an observation window for a facility. Defaults to
        ``start`` and ``end``.
        """
        return 'start', 'end'

    def get_terminal_observing_states(self):
        """
        Returns the states for which an observation is not expected
        to change.
        """
        return EXAMPLE_TERMINAL_OBSERVING_STATES

    def get_observing_sites(self):
        """
        Return a list of dictionaries that contain the information
        necessary to be used in the planning (visibility) tool. The
        list should contain dictionaries each that contain sitecode,
        latitude, longitude and elevation.
        """
        return EXAMPLE_SITES

    def data_products(self, observation_id, product_id=None):
        """
        Using an observation_id, retrieve a list of the data
        products that belong to this observation. In this case,
        the LCO module retrieves a list of frames from the LCO
        data archive.
        """
        return []

    def get_observation_url(self, observation_id):
        return ''



import requests

from django import forms

from tom_alerts.alerts import GenericQueryForm, GenericAlert, GenericBroker
from tom_targets.models import Target

LASAIR_URL = 'https://lasair.roe.ac.uk'


class LasairBrokerForm(GenericQueryForm):
    name = forms.CharField(required=True)
    cone = forms.CharField(required=False, label='Object Cone Search', help_text='Object RA and Dec')
    sqlquery = forms.CharField(required=False, label='Freeform SQL query', help_text='SQL query')

    def clean(self):
        cleaned_data = super().clean()

        # Ensure that either cone search or sqlquery are populated
        if not (cleaned_data['cone'] or cleaned_data['sqlquery']):
            raise forms.ValidationError('One of either Object Cone Search or Freeform SQL Query must be populated.')

        return cleaned_data


def get_lasair_object(objectId):
    url = LASAIR_URL + '/object/' + objectId + '/json/'
    response = requests.get(url)
    obj = response.json()
    jdmax = obj['candidates'][0]['mjd']
    ra = obj['objectData']['ramean']
    dec = obj['objectData']['decmean']
    glon = obj['objectData']['glonmean']
    glat = obj['objectData']['glatmean']
    magpsf = obj['candidates'][0]['magpsf']
    return {
        'alert_id': objectId,
        'timestamp': jdmax,
        'ra': ra,
        'dec': dec,
        'galactic_lng': glon,
        'galactic_lat': glat,
        'mag': magpsf
    }


class LasairBroker(GenericBroker):
    """
    The ``LasairBroker`` is the interface to the Lasair alert broker. For information regarding the query format for
    Lasair, please see https://lasair.roe.ac.uk/objlist/.
    """

    name = 'Lasair'
    form = LasairBrokerForm

    def fetch_alerts(self, parameters):
        if 'cone' in parameters and len(parameters['cone'].strip()) > 0:
            response = requests.post(
                LASAIR_URL + '/conesearch/',
                data={'cone': parameters['cone'], 'json': 'on'}
            )
            cone_result = response.json()
            alerts = []
            for objectId in cone_result['hitlist']:
                alerts.append(get_lasair_object(objectId))
            return iter(alerts)

        # note: the sql SELECT must include objectId
        if 'sqlquery' in parameters and len(parameters['sqlquery'].strip()) > 0:
            response = requests.post(
                LASAIR_URL + '/objlist/',
                data={'sqlquery': parameters['sqlquery'], 'json': 'on', 'page': ''}
            )
            records = response.json()
            alerts = []
            for record in records:
                alerts.append(get_lasair_object(record['objectId']))
            return iter(alerts)

    def fetch_alert(self, alert_id):
        url = LASAIR_URL + '/object/' + alert_id + '/json/'
        response = requests.get(url)
        response.raise_for_status()
        parsed = response.json()
        return parsed

    def process_reduced_data(self, target, alert=None):
        pass

    def to_generic_alert(self, alert):
        return GenericAlert(
            url=LASAIR_URL + '/object/' + alert['alert_id'],
            id=alert['alert_id'],
            name=alert['alert_id'],
            ra=alert['ra'],
            dec=alert['dec'],
            timestamp=alert['timestamp'],
            mag=alert['mag'],
            score=1,  # dunno what this means ..?
        )

    def to_target(self, alert):
        for c in alert['candidates']:
            if 'candid' in c:
                break
        return Target.objects.create(
            name=alert.get('objectId'),
            type='SIDEREAL',
            ra=alert['objectData']['ramean'],
            dec=alert['objectData']['decmean'],
            galactic_lng=alert['objectData']['glonmean'],
            galactic_lat=alert['objectData']['glatmean'],
        )



# Generated by Django 2.2.9 on 2020-02-04 23:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_observations', '0006_observationgroup_cadence_strategy'),
    ]

    operations = [
        migrations.CreateModel(
            name='ObservingStrategy',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
                ('facility', models.CharField(max_length=50)),
                ('parameters', models.TextField()),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
            ],
        ),
    ]



# Generated by Django 3.1.5 on 2021-02-04 23:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_alerts', '0003_auto_20190701_2226'),
    ]

    operations = [
        migrations.AlterField(
            model_name='brokerquery',
            name='parameters',
            field=models.JSONField(),
        ),
    ]



#!/usr/bin/env python
import os
import sys

if __name__ == "__main__":
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tom_base.settings")
    try:
        from django.core.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError(
            "Couldn't import Django. Are you sure it's installed and "
            "available on your PYTHONPATH environment variable? Did you "
            "forget to activate a virtual environment?"
        )
    execute_from_command_line(sys.argv)



from urllib.parse import urlencode

from django import template
from django.conf import settings
from django.contrib.auth.models import Group
from django.core.paginator import Paginator
from django.shortcuts import reverse
from datetime import datetime
from guardian.shortcuts import get_objects_for_user
from plotly import offline
import plotly.graph_objs as go

from tom_dataproducts.forms import DataProductUploadForm
from tom_dataproducts.models import DataProduct, ReducedDatum
from tom_dataproducts.processors.data_serializers import SpectrumSerializer
from tom_observations.models import ObservationRecord
from tom_targets.models import Target

register = template.Library()


@register.inclusion_tag('tom_dataproducts/partials/dataproduct_list_for_target.html', takes_context=True)
def dataproduct_list_for_target(context, target):
    """
    Given a ``Target``, returns a list of ``DataProduct`` objects associated with that ``Target``
    """
    if settings.TARGET_PERMISSIONS_ONLY:
        target_products_for_user = target.dataproduct_set.all()
    else:
        target_products_for_user = get_objects_for_user(
            context['request'].user, 'tom_dataproducts.view_dataproduct', klass=target.dataproduct_set.all())
    return {
        'products': target_products_for_user,
        'target': target
    }


@register.inclusion_tag('tom_dataproducts/partials/saved_dataproduct_list_for_observation.html')
def dataproduct_list_for_observation_saved(data_products, request):
    """
    Given a dictionary of dataproducts from an ``ObservationRecord``, returns the subset that are saved to the TOM. This
    templatetag paginates the subset of ``DataProduct``, and therefore requires the request to have a 'page_saved' key.

    This templatetag is intended to be used with the ``all_data_products()`` method from a facility, as it returns a
    dictionary with keys of ``saved`` and ``unsaved`` that have values of lists of ``DataProduct`` objects.
    """
    page = request.GET.get('page_saved')
    paginator = Paginator(data_products['saved'], 25)
    products_page = paginator.get_page(page)
    return {'products_page': products_page}


@register.inclusion_tag('tom_dataproducts/partials/unsaved_dataproduct_list_for_observation.html')
def dataproduct_list_for_observation_unsaved(data_products):
    """
    Given a dictionary of dataproducts from an ``ObservationRecord``, returns a list of the subset that are not saved to
    the TOM.

    This templatetag is intended to be used with the ``all_data_products()`` method from a facility, as it returns a
    dictionary with keys of ``saved`` and ``unsaved`` that have values of lists of ``DataProduct`` objects.
    """
    return {'products': data_products['unsaved']}


@register.inclusion_tag('tom_dataproducts/partials/dataproduct_list.html', takes_context=True)
def dataproduct_list_all(context):
    """
    Returns the full list of data products in the TOM, with the most recent first.
    """
    if settings.TARGET_PERMISSIONS_ONLY:
        products = DataProduct.objects.all().order_by('-created')
    else:
        products = get_objects_for_user(context['request'].user, 'tom_dataproducts.view_dataproduct')
    return {'products': products}


@register.inclusion_tag('tom_dataproducts/partials/upload_dataproduct.html', takes_context=True)
def upload_dataproduct(context, obj):
    user = context['user']
    initial = {}
    if isinstance(obj, Target):
        initial['target'] = obj
        initial['referrer'] = reverse('tom_targets:detail', args=(obj.id,))
    elif isinstance(obj, ObservationRecord):
        initial['observation_record'] = obj
        initial['referrer'] = reverse('tom_observations:detail', args=(obj.id,))
    form = DataProductUploadForm(initial=initial)
    if not settings.TARGET_PERMISSIONS_ONLY:
        if user.is_superuser:
            form.fields['groups'].queryset = Group.objects.all()
        else:
            form.fields['groups'].queryset = user.groups.all()
    return {'data_product_form': form}


@register.inclusion_tag('tom_dataproducts/partials/recent_photometry.html')
def recent_photometry(target, limit=1):
    """
    Displays a table of the most recent photometric points for a target.
    """
    photometry = ReducedDatum.objects.filter(data_type='photometry', target=target).order_by('-timestamp')[:limit]
    return {'data': [{'timestamp': rd.timestamp, 'magnitude': rd.value['magnitude']} for rd in photometry]}


@register.inclusion_tag('tom_dataproducts/partials/photometry_for_target.html', takes_context=True)
def photometry_for_target(context, target, width=700, height=600, background=None, label_color=None, grid=True):
    """
    Renders a photometric plot for a target.

    This templatetag requires all ``ReducedDatum`` objects with a data_type of ``photometry`` to be structured with the
    following keys in the JSON representation: magnitude, error, filter

    :param width: Width of generated plot
    :type width: int

    :param height: Height of generated plot
    :type width: int

    :param background: Color of the background of generated plot. Can be rgba or hex string.
    :type background: str

    :param label_color: Color of labels/tick labels. Can be rgba or hex string.
    :type label_color: str

    :param grid: Whether to show grid lines.
    :type grid: bool
    """

    color_map = {
        'r': 'red',
        'g': 'green',
        'i': 'black'
    }

    photometry_data = {}
    if settings.TARGET_PERMISSIONS_ONLY:
        datums = ReducedDatum.objects.filter(target=target, data_type=settings.DATA_PRODUCT_TYPES['photometry'][0])
    else:
        datums = get_objects_for_user(context['request'].user,
                                      'tom_dataproducts.view_reduceddatum',
                                      klass=ReducedDatum.objects.filter(
                                        target=target,
                                        data_type=settings.DATA_PRODUCT_TYPES['photometry'][0]))

    for datum in datums:
        photometry_data.setdefault(datum.value['filter'], {})
        photometry_data[datum.value['filter']].setdefault('time', []).append(datum.timestamp)
        photometry_data[datum.value['filter']].setdefault('magnitude', []).append(datum.value.get('magnitude'))
        photometry_data[datum.value['filter']].setdefault('error', []).append(datum.value.get('error'))
        photometry_data[datum.value['filter']].setdefault('limit', []).append(datum.value.get('limit'))

    plot_data = []
    for filter_name, filter_values in photometry_data.items():
        if filter_values['magnitude']:
            series = go.Scatter(
                x=filter_values['time'],
                y=filter_values['magnitude'],
                mode='markers',
                marker=dict(color=color_map.get(filter_name)),
                name=filter_name,
                error_y=dict(
                    type='data',
                    array=filter_values['error'],
                    visible=True
                )
            )
            plot_data.append(series)
        if filter_values['limit']:
            series = go.Scatter(
                x=filter_values['time'],
                y=filter_values['limit'],
                mode='markers',
                opacity=0.5,
                marker=dict(color=color_map.get(filter_name)),
                marker_symbol=6,  # upside down triangle
                name=filter_name + ' non-detection',
            )
            plot_data.append(series)

    layout = go.Layout(
        yaxis=dict(autorange='reversed'),
        height=height,
        width=width,
        paper_bgcolor=background,
        plot_bgcolor=background

    )
    layout.legend.font.color = label_color
    fig = go.Figure(data=plot_data, layout=layout)
    fig.update_yaxes(showgrid=grid, color=label_color, showline=True, linecolor=label_color, mirror=True)
    fig.update_xaxes(showgrid=grid, color=label_color, showline=True, linecolor=label_color, mirror=True)
    return {
        'target': target,
        'plot': offline.plot(fig, output_type='div', show_link=False)
    }


@register.inclusion_tag('tom_dataproducts/partials/spectroscopy_for_target.html', takes_context=True)
def spectroscopy_for_target(context, target, dataproduct=None):
    """
    Renders a spectroscopic plot for a ``Target``. If a ``DataProduct`` is specified, it will only render a plot with
    that spectrum.
    """
    spectral_dataproducts = DataProduct.objects.filter(target=target,
                                                       data_product_type=settings.DATA_PRODUCT_TYPES['spectroscopy'][0])
    if dataproduct:
        spectral_dataproducts = DataProduct.objects.get(data_product=dataproduct)

    plot_data = []
    if settings.TARGET_PERMISSIONS_ONLY:
        datums = ReducedDatum.objects.filter(data_product__in=spectral_dataproducts)
    else:
        datums = get_objects_for_user(context['request'].user,
                                      'tom_dataproducts.view_reduceddatum',
                                      klass=ReducedDatum.objects.filter(data_product__in=spectral_dataproducts))
    for datum in datums:
        deserialized = SpectrumSerializer().deserialize(datum.value)
        plot_data.append(go.Scatter(
            x=deserialized.wavelength.value,
            y=deserialized.flux.value,
            name=datetime.strftime(datum.timestamp, '%Y%m%d-%H:%M:%s')
        ))

    layout = go.Layout(
        height=600,
        width=700,
        xaxis=dict(
            tickformat="d"
        ),
        yaxis=dict(
            tickformat=".1eg"
        )
    )
    return {
        'target': target,
        'plot': offline.plot(go.Figure(data=plot_data, layout=layout), output_type='div', show_link=False)
    }


@register.inclusion_tag('tom_dataproducts/partials/update_broker_data_button.html', takes_context=True)
def update_broker_data_button(context):
    return {'query_params': urlencode(context['request'].GET.dict())}



import json
from unittest.mock import patch

from django import forms
from django.contrib.auth.models import User, Group
from django.contrib.messages import get_messages
from django.core.cache import cache
from django.test import TestCase, override_settings
from django.urls import reverse

from tom_alerts.alerts import GenericBroker, GenericQueryForm, GenericUpstreamSubmissionForm, GenericAlert
from tom_alerts.alerts import get_service_class, get_service_classes
from tom_alerts.exceptions import AlertSubmissionException
from tom_alerts.models import BrokerQuery
from tom_observations.models import ObservationRecord
from tom_targets.models import Target

# Test alert data. Normally this would come from a remote source.
test_alerts = [
    {'id': 1, 'name': 'Tatooine', 'timestamp': '2019-07-01', 'ra': 32, 'dec': -20, 'mag': 8, 'score': 20},
    {'id': 2, 'name': 'Hoth', 'timestamp': '2019-07-02', 'ra': 66, 'dec': 50, 'mag': 3, 'score': 66},
]


class TestBrokerForm(GenericQueryForm):
    """ All brokers must have a form which will be used to construct and save queries
    to the broker. They should subclass `GenericQueryForm` which includes some required
    fields and contains logic for serializing and persisting the query parameters to the
    database. This test form will only have one field.
    """
    name = forms.CharField(required=True)


class TestUpstreamSubmissionForm(GenericUpstreamSubmissionForm):
    """
    Brokers supporting upstream submission can have a form used for constructing the submission. If should subclass
    GenericUpstreamSubmissionForm. This test form will have only one additional field in order to test that the
    additional field value is submitted to the broker correctly.
    """
    topic = forms.CharField(required=False)


class TestBroker(GenericBroker):
    """ The broker class encapsulates the logic for querying remote brokers and transforming
    the returned data into TOM Toolkit Targets so they can be used elsewhere in the system. The
    following methods and attributes are all required, but a broker can be as complex as needed.
    """
    name = 'TEST'  # The name of this broker.
    form = TestBrokerForm  # The form that will be used to write and save queries.
    alert_submission_form = TestUpstreamSubmissionForm

    def fetch_alerts(self, parameters):
        """ All brokers must implement this method. It must return a list of alerts.
        """
        # Here we simply return a list of `GenericAlert`s that match the name passed in via `parameters`.
        return iter([alert for alert in test_alerts if alert['name'] == parameters['name']])

    def process_reduced_data(self, target, alert=None):
        pass

    def to_generic_alert(self, alert):
        """ We use this method to transform a remote alert (in this case an item
        from the `test_alerts` list) into a `GenericAlert` so they can be displayed in a
        consistent manner.
        """
        return GenericAlert(
            timestamp=alert['timestamp'],
            url='https://tomtoolkit.github.io',
            id=alert['id'],
            name=alert['name'],
            ra=alert['ra'],
            dec=alert['dec'],
            mag=alert['mag'],
            score=alert['score']
        )

    def submit_upstream_alert(self, target=None, observation_record=None, **kwargs):
        return super().submit_upstream_alert(target=target, observation_record=observation_record)


@override_settings(TOM_ALERT_CLASSES=['tom_alerts.tests.tests.TestBroker'])
class TestBrokerClass(TestCase):
    """ Test the functionality of the TestBroker, we modify the django settings to make sure
    it is the only installed broker.
    """
    def test_get_broker_class(self):
        self.assertEqual(TestBroker, get_service_class('TEST'))

    def test_get_invalid_broker(self):
        with self.assertRaises(ImportError):
            get_service_class('MARS')

    def test_fetch_alerts(self):
        alerts = TestBroker().fetch_alerts({'name': 'Hoth'})
        self.assertEqual(test_alerts[1], list(alerts)[0])

    def test_to_generic_alert(self):
        ga = TestBroker().to_generic_alert(test_alerts[0])
        self.assertEqual(ga.name, test_alerts[0]['name'])

    def test_to_target(self):
        target, _, _ = TestBroker().to_generic_alert(test_alerts[0]).to_target()
        self.assertEqual(target.name, test_alerts[0]['name'])


@override_settings(TOM_ALERT_CLASSES=['tom_alerts.fake_broker'])
class TestAlertModule(TestCase):
    """Test that attempting to import a nonexistent broker module raises the appropriate errors.
    """

    def test_get_service_classes_import_error(self):
        with self.subTest('Invalid import returns an import error.'):
            with patch('tom_alerts.alerts.import_module') as mock_import_module:
                mock_import_module.side_effect = ImportError()
                with self.assertRaisesRegex(ImportError, 'Could not import tom_alerts.fake_broker.'):
                    get_service_classes()

        with self.subTest('Invalid import returns an attribute error.'):
            with patch('tom_alerts.alerts.import_module') as mock_import_module:
                mock_import_module.side_effect = AttributeError()
                with self.assertRaisesRegex(ImportError, 'Could not import tom_alerts.fake_broker.'):
                    get_service_classes()


@override_settings(TOM_ALERT_CLASSES=['tom_alerts.tests.tests.TestBroker'])
class TestBrokerViews(TestCase):
    """ Test the views that use the broker classes
    """
    def setUp(self):
        self.user = User.objects.create(username='Han', email='han@example.com')
        group = Group.objects.create(name='test')
        group.user_set.add(self.user)
        group.save()
        self.client.force_login(self.user)

    def test_display_form(self):
        response = self.client.get(reverse('tom_alerts:create') + '?broker=TEST')
        self.assertContains(response, 'TEST Query Form')

    def test_create_query(self):
        query_data = {
            'query_name': 'Test Query',
            'broker': 'TEST',
            'name': 'Test Name',
        }
        response = self.client.post(reverse('tom_alerts:create'), data=query_data, follow=True)
        self.assertContains(response, query_data['query_name'])
        self.assertEqual(BrokerQuery.objects.count(), 1)
        self.assertEqual(BrokerQuery.objects.first().name, query_data['query_name'])

    def test_filter_queries(self):
        broker_query = BrokerQuery.objects.create(
            name='Is it dust?',
            broker='TEST',
            parameters={'name': 'Alderaan'},
        )
        not_found = BrokerQuery.objects.create(
            name='find hoth',
            broker='TEST',
            parameters={'name': 'Hoth'},
        )
        response = self.client.get(reverse('tom_alerts:list') + '?name=dust')
        self.assertContains(response, broker_query.name)
        self.assertNotContains(response, not_found.name)

    def test_delete_query(self):
        broker_query = BrokerQuery.objects.create(
            name='find hoth',
            broker='TEST',
            parameters={'name': 'Hoth'},
        )
        self.assertTrue(BrokerQuery.objects.filter(name='find hoth').exists())
        self.client.post(reverse('tom_alerts:delete', kwargs={'pk': broker_query.id}))
        self.assertFalse(BrokerQuery.objects.filter(name='find hoth').exists())

    def test_run_query(self):
        broker_query = BrokerQuery.objects.create(
            name='find hoth',
            broker='TEST',
            parameters={'name': 'Hoth'},
        )
        response = self.client.get(reverse('tom_alerts:run', kwargs={'pk': broker_query.id}))
        self.assertContains(response,  '66')

    def test_update_query(self):
        broker_query = BrokerQuery.objects.create(
            name='find hoth',
            broker='TEST',
            parameters={'name': 'Hoth'},
        )
        update_data = {
            'query_name': 'find hoth',
            'broker': 'TEST',
            'name': 'another place',
        }
        self.client.post(reverse('tom_alerts:update', kwargs={'pk': broker_query.id}), data=update_data)
        broker_query.refresh_from_db()
        self.assertEqual(broker_query.parameters['name'], update_data['name'])

    @override_settings(CACHES={
            'default': {
                'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
            }
        })
    def test_create_target(self):
        cache.set('alert_2', json.dumps(test_alerts[1]))
        query = BrokerQuery.objects.create(
            name='find hoth',
            broker='TEST',
            parameters={'name': 'Hoth'},
        )
        post_data = {
            'broker': 'TEST',
            'query_id': query.id,
            'alerts': [2]
        }
        response = self.client.post(reverse('tom_alerts:create-target'), data=post_data)
        self.assertEqual(Target.objects.count(), 1)
        self.assertEqual(Target.objects.first().name, 'Hoth')
        self.assertRedirects(response, reverse('tom_targets:update', kwargs={'pk': Target.objects.first().id}))

    @override_settings(CACHES={
            'default': {
                'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
            }
        })
    def test_create_multiple_targets(self):
        cache.set('alert_1', json.dumps(test_alerts[0]))
        cache.set('alert_2', json.dumps(test_alerts[1]))
        query = BrokerQuery.objects.create(
            name='find anything',
            broker='TEST',
            parameters={'score__gt': 19},
        )
        post_data = {
            'broker': 'TEST',
            'query_id': query.id,
            'alerts': [1, 2]
        }
        response = self.client.post(reverse('tom_alerts:create-target'), data=post_data)
        self.assertEqual(Target.objects.count(), 2)
        self.assertRedirects(response, reverse('tom_targets:list'))

    def test_create_no_targets(self):
        query = BrokerQuery.objects.create(
            name='find anything',
            broker='TEST',
            parameters={'name': 'Alderaan'},
        )
        post_data = {
            'broker': 'TEST',
            'query_id': query.id,
            'alerts': []
        }
        response = self.client.post(reverse('tom_alerts:create-target'), data=post_data, follow=True)
        self.assertEqual(Target.objects.count(), 0)
        self.assertRedirects(response, reverse('tom_alerts:run', kwargs={'pk': query.id}))

    @patch('tom_alerts.tests.tests.TestBroker.submit_upstream_alert')
    def test_submit_alert_success(self, mock_submit_upstream_alert):
        """Test submission of an alert to a broker."""

        # Tests that an alert is submitted with just a target, and that no redirect_url results in redirect to home
        target = Target.objects.create(name='test_target', ra=1, dec=2)
        response = self.client.post(reverse('tom_alerts:submit-alert', kwargs={'broker': 'TEST'}),
                                    data={'target': target.id})

        mock_submit_upstream_alert.assert_called_with(target=target, observation_record=None, topic='')
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse('home'))

        # Tests that an alert is submitted with just an observation, and that redirect is to redirect_url
        obsr = ObservationRecord.objects.create(target=target, facility='Test', parameters={}, observation_id=1)
        response = self.client.post(reverse('tom_alerts:submit-alert', kwargs={'broker': 'TEST'}),
                                    data={'observation_record': obsr.id, 'redirect_url': reverse('tom_targets:list')})

        mock_submit_upstream_alert.assert_called_with(target=None, observation_record=obsr, topic='')
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse('tom_targets:list'))

        # Tests that an alert submitted with additional parameters calls submit_upstream_alert correctly.
        response = self.client.post(reverse('tom_alerts:submit-alert', kwargs={'broker': 'TEST'}),
                                    data={'observation_record': obsr.id, 'topic': 'test topic'})
        mock_submit_upstream_alert.assert_called_with(target=None, observation_record=obsr, topic='test topic')

    @patch('tom_alerts.tests.tests.TestBroker.submit_upstream_alert')
    def test_submit_alert_failure(self, mock_submit_upstream_alert):
        """Test that a failed alert submission returns an appropriate message."""
        target = Target.objects.create(name='test_target', ra=1, dec=2)
        mock_submit_upstream_alert.return_value = False
        response = self.client.post(reverse('tom_alerts:submit-alert', kwargs={'broker': 'TEST'}),
                                    data={'target': target.id})
        messages = [(m.message, m.level) for m in get_messages(response.wsgi_request)]
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0][0], 'Unable to submit one or more alerts to TEST. See logs for details.')

    @patch('tom_alerts.tests.tests.TestBroker.submit_upstream_alert')
    def test_submit_alert_exception(self, mock_submit_upstream_alert):
        """Test that an alert submission returns an appropriate message when alert submission raises an exception."""
        mock_submit_upstream_alert.side_effect = AlertSubmissionException()

        response = self.client.post(reverse('tom_alerts:submit-alert', kwargs={'broker': 'TEST'}), data={})
        messages = [(m.message, m.level) for m in get_messages(response.wsgi_request)]
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0][0], 'Unable to submit one or more alerts to TEST. See logs for details.')

    def test_submit_alert_invalid_form(self):
        """Test that an alert submission failed when form is invalid."""
        response = self.client.post(reverse('tom_alerts:submit-alert', kwargs={'broker': 'TEST'}), data={})
        messages = [(m.message, m.level) for m in get_messages(response.wsgi_request)]
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0][0], 'Unable to submit one or more alerts to TEST. See logs for details.')
        self.assertRedirects(response, reverse('home'))



from django.urls import path

from tom_alerts.views import BrokerQueryCreateView, BrokerQueryListView, BrokerQueryUpdateView, BrokerQueryDeleteView
from tom_alerts.views import CreateTargetFromAlertView, RunQueryView, SubmitAlertUpstreamView

app_name = 'tom_alerts'

urlpatterns = [
    path('query/list/', BrokerQueryListView.as_view(), name='list'),
    path('query/create/', BrokerQueryCreateView.as_view(), name='create'),
    path('query/<int:pk>/update/', BrokerQueryUpdateView.as_view(), name='update'),
    path('query/<int:pk>/run/', RunQueryView.as_view(), name='run'),
    path('query/<int:pk>/delete/', BrokerQueryDeleteView.as_view(), name='delete'),
    path('alert/create/', CreateTargetFromAlertView.as_view(), name='create-target'),
    path('<str:broker>/submit/', SubmitAlertUpstreamView.as_view(), name='submit-alert')
]



from tom_catalogs.harvester import AbstractHarvester

from astroquery.mpc import MPC


class MPCHarvester(AbstractHarvester):
    """
    The ``MPCHarvester`` is the interface to the Minor Planet Center catalog. For information regarding the Minor Planet
    Center catalog, please see https://minorplanetcenter.net/ or
    https://astroquery.readthedocs.io/en/latest/mpc/mpc.html.
    """

    name = 'MPC'

    def query(self, term):
        self.catalog_data = MPC.query_object('asteroid', name=term)

    def to_target(self):
        target = super().to_target()
        result = self.catalog_data[0]
        target.type = 'NON_SIDEREAL'
        target.name = result['name']
        target.extra_names = [result['designation']] if result['designation'] else []
        target.epoch_of_elements = self.jd_to_mjd(result['epoch_jd'])
        target.mean_anomaly = result['mean_anomaly']
        target.arg_of_perihelion = result['argument_of_perihelion']
        target.eccentricity = result['eccentricity']
        target.lng_asc_node = result['ascending_node']
        target.inclination = result['inclination']
        target.mean_daily_motion = result['mean_daily_motion']
        target.semimajor_axis = result['semimajor_axis']
        return target



from django.urls import path

from tom_dataproducts.views import DataProductListView, DataProductSaveView, DataProductGroupListView
from tom_dataproducts.views import DataProductDeleteView, DataProductGroupCreateView
from tom_dataproducts.views import DataProductGroupDetailView, DataProductGroupDataView, DataProductGroupDeleteView
from tom_dataproducts.views import DataProductUploadView, DataProductFeatureView, UpdateReducedDataView

from tom_common.api_router import SharedAPIRootRouter
from tom_dataproducts.api_views import DataProductViewSet

router = SharedAPIRootRouter()
router.register(r'dataproducts', DataProductViewSet, 'dataproducts')

app_name = 'tom_dataproducts'

urlpatterns = [
    path('data/', DataProductListView.as_view(), name='list'),
    path('data/group/create/', DataProductGroupCreateView.as_view(), name='group-create'),
    path('data/group/list/', DataProductGroupListView.as_view(), name='group-list'),
    path('data/group/add/', DataProductGroupDataView.as_view(), name='group-data'),
    path('data/group/<int:pk>/', DataProductGroupDetailView.as_view(), name='group-detail'),
    path('data/group/<int:pk>/delete/', DataProductGroupDeleteView.as_view(), name='group-delete'),
    path('data/upload/', DataProductUploadView.as_view(), name='upload'),
    path('data/reduced/update/', UpdateReducedDataView.as_view(), name='update-reduced-data'),
    path('data/<int:pk>/delete/', DataProductDeleteView.as_view(), name='delete'),
    path('data/<int:pk>/feature/', DataProductFeatureView.as_view(), name='feature'),
    path('<int:pk>/save/', DataProductSaveView.as_view(), name='save'),
]



# Generated by Django 3.1.5 on 2021-08-11 00:18

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tom_targets', '0018_auto_20200714_1832'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='target',
            options={},
        ),
    ]



from unittest import mock

from django.test import override_settings, tag, TestCase

from tom_alerts.alerts import get_service_class
from tom_alerts.brokers.lasair import LasairBroker, LasairBrokerForm

alert1 = {
    'candidates':
    [
        {
            'decl': 30.5462915,
            'magpsf': 15.4319,
            'ra': 52.6833968,
            'candid': 1638483221215015008,
        },
        {
            'decl': 30.5462627,
            'magpsf': 16.5676,
            'ra': 52.6833575,
            'candid': 1649478961215015007,
        }
    ],
    'objectData': {
            'ramean': 52.68339748783069,
            'decmean': 30.546314428042333,
            'glonmean': 159.12011347419698,
            'glatmean': -20.911037347147648,
    },
    'objectId': 'ZTF18aaaduje',
}


class TestLasairBrokerForm(TestCase):
    def setUp(self):
        pass

    def test_clean(self):
        form_parameters = {'query_name': 'Test Lasair', 'broker': 'Lasair', 'name': 'ZTF18abbkloa',
                           'cone': '', 'sqlquery': ''}

        with self.subTest():
            form = LasairBrokerForm(form_parameters)
            self.assertFalse(form.is_valid())
            self.assertIn('One of either Object Cone Search or Freeform SQL Query must be populated.',
                          form.non_field_errors())

        test_parameters_list = [{'cone': '1, 2', 'sqlquery': ''}, {'cone': '', 'sqlquery': 'select * from objects;'}]
        for test_params in test_parameters_list:
            with self.subTest():
                form_parameters.update(test_params)
                form = LasairBrokerForm(form_parameters)
                self.assertTrue(form.is_valid())


@override_settings(TOM_ALERT_CLASSES=['tom_alerts.brokers.lasair.LasairBroker'])
class TestLasairBrokerClass(TestCase):

    """ Test the functionality of the LasairBroker, we modify the django settings to make sure
    it is the only installed broker.
    """

    def setUp(self):
        pass

    def test_get_broker_class(self):
        self.assertEqual(LasairBroker, get_service_class('Lasair'))

    @mock.patch('tom_alerts.brokers.lasair.requests.get')
    def test_fetch_alerts(self, mock_requests_get):
        pass

    def test_to_target(self):
        created_target = LasairBroker().to_target(alert1)  # uses to_target method
        '''
        The following assertions check to_target() works properly
        '''
        self.assertEqual(created_target.name, 'ZTF18aaaduje')
        self.assertEqual(created_target.ra, 52.68339748783069)
        self.assertEqual(created_target.dec, 30.546314428042333)
        self.assertEqual(created_target.galactic_lng, 159.12011347419698)
        self.assertEqual(created_target.galactic_lat, -20.911037347147648)

    def test_to_generic_alert(self):
        pass


@tag('canary')
class TestLasairModuleCanary(TestCase):
    def setUp(self):
        self.broker = LasairBroker()

    def test_fetch_alerts(self):
        pass

    def test_fetch_alert(self):
        pass



from django.contrib.auth.models import Group
from rest_framework import serializers


class GroupSerializer(serializers.ModelSerializer):
    id = serializers.IntegerField(required=False)
    name = serializers.CharField(required=False)

    class Meta:
        model = Group
        fields = ('id', 'name',)



# Generated by Django 2.1.4 on 2019-05-03 23:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_targets', '0006_auto_20190403_1659'),
    ]

    operations = [
        migrations.AlterField(
            model_name='target',
            name='epoch_of_perihelion',
            field=models.FloatField(blank=True, help_text='Julian Date.', null=True, verbose_name='Epoch of Perihelion'),
        ),
    ]



# Generated by Django 2.2.8 on 2019-12-10 00:07

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tom_observations', '0004_observationgroup'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='observationgroup',
            options={'ordering': ('-created',)},
        ),
    ]



from django.urls import path

from .views import TargetCreateView, TargetUpdateView, TargetDetailView, TargetNameSearchView
from .views import TargetDeleteView, TargetListView, TargetImportView, TargetExportView
from .views import TargetGroupingView, TargetGroupingDeleteView, TargetGroupingCreateView, TargetAddRemoveGroupingView

from .api_views import TargetViewSet, TargetExtraViewSet, TargetNameViewSet
from tom_common.api_router import SharedAPIRootRouter

router = SharedAPIRootRouter()
router.register(r'targets', TargetViewSet, 'targets')
router.register(r'targetextra', TargetExtraViewSet, 'targetextra')
router.register(r'targetname', TargetNameViewSet, 'targetname')

app_name = 'tom_targets'

urlpatterns = [
    path('', TargetListView.as_view(), name='list'),
    path('targetgrouping/', TargetGroupingView.as_view(), name='targetgrouping'),
    path('create/', TargetCreateView.as_view(), name='create'),
    path('import/', TargetImportView.as_view(), name='import'),
    path('export/', TargetExportView.as_view(), name='export'),
    path('add-remove-grouping/', TargetAddRemoveGroupingView.as_view(), name='add-remove-grouping'),
    path('name/<str:name>', TargetNameSearchView.as_view(), name='name-search'),
    path('<int:pk>/update/', TargetUpdateView.as_view(), name='update'),
    path('<int:pk>/delete/', TargetDeleteView.as_view(), name='delete'),
    path('<int:pk>/', TargetDetailView.as_view(), name='detail'),
    path('targetgrouping/<int:pk>/delete/', TargetGroupingDeleteView.as_view(), name='delete-group'),
    path('targetgrouping/create/', TargetGroupingCreateView.as_view(), name='create-group')
]



# Generated by Django 2.1.4 on 2018-12-06 21:57

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('tom_targets', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ObservationRecord',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('facility', models.CharField(max_length=50)),
                ('parameters', models.TextField()),
                ('observation_id', models.CharField(max_length=2000)),
                ('status', models.CharField(max_length=200)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('target', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tom_targets.Target')),
            ],
            options={
                'ordering': ('-created',),
            },
        ),
    ]



# Generated by Django 2.2 on 2019-12-05 19:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_dataproducts', '0007_manual_20191016_rename_type'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='reduceddatum',
            options={'get_latest_by': ('timestamp',)},
        ),
        migrations.AlterField(
            model_name='dataproduct',
            name='data_product_type',
            field=models.CharField(blank=True, default='', max_length=50),
        ),
        migrations.AlterField(
            model_name='reduceddatum',
            name='data_type',
            field=models.CharField(default='', max_length=100),
        ),
    ]



from django.utils.decorators import method_decorator
from django.contrib.auth.decorators import user_passes_test
from guardian.mixins import PermissionRequiredMixin


class SuperuserRequiredMixin():
    @method_decorator(user_passes_test(lambda u: u.is_superuser))
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)


class Raise403PermissionRequiredMixin(PermissionRequiredMixin):
    return_403 = True



from rest_framework.mixins import ListModelMixin
from rest_framework.viewsets import GenericViewSet

from tom_common.serializers import GroupSerializer


class GroupViewSet(ListModelMixin, GenericViewSet):
    """
    Viewset for Group objects. By default supports list operation only.
    See the docs on viewsets: https://www.django-rest-framework.org/api-guide/viewsets/
    """
    serializer_class = GroupSerializer

    def get_queryset(self):
        return self.request.user.groups.all()



import mimetypes

from django.conf import settings
from importlib import import_module

from tom_dataproducts.models import ReducedDatum


DEFAULT_DATA_PROCESSOR_CLASS = 'tom_dataproducts.data_processor.DataProcessor'


def run_data_processor(dp):
    """
    Reads the `data_product_type` from the dp parameter and imports the corresponding `DataProcessor` specified in
    `settings.py`, then runs `process_data` and inserts the returned values into the database.

    :param dp: DataProduct which will be processed into a list
    :type dp: DataProduct

    :returns: QuerySet of `ReducedDatum` objects created by the `run_data_processor` call
    :rtype: `QuerySet` of `ReducedDatum`
    """

    try:
        processor_class = settings.DATA_PROCESSORS[dp.data_product_type]
    except Exception:
        processor_class = DEFAULT_DATA_PROCESSOR_CLASS

    try:
        mod_name, class_name = processor_class.rsplit('.', 1)
        mod = import_module(mod_name)
        clazz = getattr(mod, class_name)
    except (ImportError, AttributeError):
        raise ImportError('Could not import {}. Did you provide the correct path?'.format(processor_class))

    data_processor = clazz()
    data = data_processor.process_data(dp)

    reduced_datums = [ReducedDatum(target=dp.target, data_product=dp, data_type=dp.data_product_type,
                                   timestamp=datum[0], value=datum[1]) for datum in data]
    ReducedDatum.objects.bulk_create(reduced_datums)

    return ReducedDatum.objects.filter(data_product=dp)


class DataProcessor():

    FITS_MIMETYPES = ['image/fits', 'application/fits']
    PLAINTEXT_MIMETYPES = ['text/plain', 'text/csv']

    mimetypes.add_type('image/fits', '.fits')
    mimetypes.add_type('image/fits', '.fz')
    mimetypes.add_type('application/fits', '.fits')
    mimetypes.add_type('application/fits', '.fz')

    def process_data(self, data_product):
        """
        Routes a photometry processing call to a method specific to a file-format. This method is expected to be
        implemented by any subclasses.

        :param data_product: DataProduct which will be processed into a list
        :type data_product: DataProduct

        :returns: python list of 2-tuples, each with a timestamp and corresponding data
        :rtype: list of 2-tuples
        """
        return []



from datetime import datetime, timedelta

from astroplan import moon_illumination
from astropy import units as u
from astropy.coordinates import Angle, get_moon, SkyCoord
from astropy.time import Time
from dateutil.parser import parse
from django import template
from django.conf import settings
from django.db.models import Q
from guardian.shortcuts import get_objects_for_user
import numpy as np
from plotly import offline
from plotly import graph_objs as go

from tom_observations.utils import get_sidereal_visibility
from tom_targets.models import Target, TargetExtra, TargetList
from tom_targets.forms import TargetVisibilityForm

register = template.Library()


@register.inclusion_tag('tom_targets/partials/recent_targets.html', takes_context=True)
def recent_targets(context, limit=10):
    """
    Displays a list of the most recently created targets in the TOM up to the given limit, or 10 if not specified.
    """
    user = context['request'].user
    return {'targets': get_objects_for_user(user, 'tom_targets.view_target').order_by('-created')[:limit]}


@register.inclusion_tag('tom_targets/partials/recently_updated_targets.html', takes_context=True)
def recently_updated_targets(context, limit=10):
    """
    Displays a list of the most recently updated targets in the TOM up to the given limit, or 10 if not specified.
    """
    user = context['request'].user
    return {'targets': get_objects_for_user(user, 'tom_targets.view_target').order_by('-modified')[:limit]}


@register.inclusion_tag('tom_targets/partials/target_feature.html')
def target_feature(target):
    """
    Displays the featured image for a target.
    """
    return {'target': target}


@register.inclusion_tag('tom_targets/partials/target_buttons.html')
def target_buttons(target):
    """
    Displays the Update and Delete buttons for a target.
    """
    return {'target': target}


@register.inclusion_tag('tom_targets/partials/target_data.html')
def target_data(target):
    """
    Displays the data of a target.
    """
    extras = {k['name']: target.extra_fields.get(k['name'], '') for k in settings.EXTRA_FIELDS if not k.get('hidden')}
    return {
        'target': target,
        'extras': extras
    }


@register.inclusion_tag('tom_targets/partials/target_unknown_statuses.html')
def target_unknown_statuses(target):
    return {
        'num_unknown_statuses': len(target.observationrecord_set.filter(Q(status='') | Q(status=None)))
    }


@register.inclusion_tag('tom_targets/partials/target_groups.html')
def target_groups(target):
    """
    Widget displaying groups this target is in and controls for modifying group association for the given target.
    """
    groups = TargetList.objects.filter(targets=target)
    return {'target': target,
            'groups': groups}


@register.inclusion_tag('tom_targets/partials/target_plan.html', takes_context=True)
def target_plan(context):
    """
    Displays form and renders plot for visibility calculation. Using this templatetag to render a plot requires that
    the context of the parent view have values for start_time, end_time, and airmass.
    """
    request = context['request']
    plan_form = TargetVisibilityForm()
    visibility_graph = ''
    if all(request.GET.get(x) for x in ['start_time', 'end_time']):
        plan_form = TargetVisibilityForm({
            'start_time': request.GET.get('start_time'),
            'end_time': request.GET.get('end_time'),
            'airmass': request.GET.get('airmass'),
            'target': context['object']
        })
        if plan_form.is_valid():
            start_time = parse(request.GET['start_time'])
            end_time = parse(request.GET['end_time'])
            if request.GET.get('airmass'):
                airmass_limit = float(request.GET.get('airmass'))
            else:
                airmass_limit = None
            visibility_data = get_sidereal_visibility(context['object'], start_time, end_time, 10, airmass_limit)
            plot_data = [
                go.Scatter(x=data[0], y=data[1], mode='lines', name=site) for site, data in visibility_data.items()
            ]
            layout = go.Layout(yaxis=dict(autorange='reversed'))
            visibility_graph = offline.plot(
                go.Figure(data=plot_data, layout=layout), output_type='div', show_link=False
            )
    return {
        'form': plan_form,
        'target': context['object'],
        'visibility_graph': visibility_graph
    }


@register.inclusion_tag('tom_targets/partials/moon_distance.html')
def moon_distance(target, day_range=30):
    """
    Renders plot for lunar distance from sidereal target.

    Adapted from Jamison Frost Burke's moon visibility code in Supernova Exchange 2.0, as seen here:
    https://github.com/jfrostburke/snex2/blob/0c1eb184c942cb10f7d54084e081d8ac11700edf/custom_code/templatetags/custom_code_tags.py#L196

    :param target: Target object for which moon distance is calculated
    :type target: tom_targets.models.Target

    :param day_range: Number of days to plot lunar distance
    :type day_range: int
    """
    if target.type != 'SIDEREAL':
        return {'plot': None}

    day_range = 30
    times = Time(
        [str(datetime.utcnow() + timedelta(days=delta)) for delta in np.arange(0, day_range, 0.2)],
        format='iso', scale='utc'
    )

    obj_pos = SkyCoord(target.ra, target.dec, unit=u.deg)
    moon_pos = get_moon(times)

    separations = moon_pos.separation(obj_pos).deg
    phases = moon_illumination(times)

    distance_color = 'rgb(0, 0, 255)'
    phase_color = 'rgb(255, 0, 0)'
    plot_data = [
        go.Scatter(x=times.mjd-times[0].mjd, y=separations, mode='lines', name='Moon distance (degrees)',
                   line=dict(color=distance_color)),
        go.Scatter(x=times.mjd-times[0].mjd, y=phases, mode='lines', name='Moon phase', yaxis='y2',
                   line=dict(color=phase_color))
    ]
    layout = go.Layout(
                xaxis={'title': 'Days from now'},
                yaxis={'range': [0, 180], 'tick0': 0, 'dtick': 45, 'tickfont': {'color': distance_color}},
                yaxis2={'range': [0, 1], 'tick0': 0, 'dtick': 0.25, 'overlaying': 'y', 'side': 'right',
                        'tickfont': {'color': phase_color}},
                margin={'l': 20, 'r': 10, 'b': 30, 't': 40},
                width=600,
                height=300,
                autosize=True
            )
    moon_distance_plot = offline.plot(
        go.Figure(data=plot_data, layout=layout), output_type='div', show_link=False
    )

    return {'plot': moon_distance_plot}


@register.inclusion_tag('tom_targets/partials/target_distribution.html')
def target_distribution(targets):
    """
    Displays a plot showing on a map the locations of all sidereal targets in the TOM.
    """
    locations = targets.filter(type=Target.SIDEREAL).values_list('ra', 'dec', 'name')
    data = [
        dict(
            lon=[location[0] for location in locations],
            lat=[location[1] for location in locations],
            text=[location[2] for location in locations],
            hoverinfo='lon+lat+text',
            mode='markers',
            type='scattergeo'
        ),
        dict(
            lon=list(range(0, 360, 60))+[180]*4,
            lat=[0]*6+[-60, -30, 30, 60],
            text=list(range(0, 360, 60))+[-60, -30, 30, 60],
            hoverinfo='none',
            mode='text',
            type='scattergeo'
        )
    ]
    layout = {
        'title': 'Target Distribution (sidereal)',
        'hovermode': 'closest',
        'showlegend': False,
        'geo': {
            'projection': {
                'type': 'mollweide',
            },
            'showcoastlines': False,
            'showland': False,
            'lonaxis': {
                'showgrid': True,
                'range': [0, 360],
            },
            'lataxis': {
                'showgrid': True,
                'range': [-90, 90],
            },
        }
    }
    figure = offline.plot(go.Figure(data=data, layout=layout), output_type='div', show_link=False)
    return {'figure': figure}


@register.filter
def deg_to_sexigesimal(value, fmt):
    """
    Displays a degree coordinate value in sexigesimal, given a format of hms or dms.
    """
    a = Angle(value, unit=u.degree)
    if fmt == 'hms':
        return '{0:02.0f}:{1:02.0f}:{2:05.3f}'.format(a.hms.h, a.hms.m, a.hms.s)
    elif fmt == 'dms':
        rep = a.signed_dms
        sign = '-' if rep.sign < 0 else '+'
        return '{0}{1:02.0f}:{2:02.0f}:{3:05.3f}'.format(sign, rep.d, rep.m, rep.s)
    else:
        return 'fmt must be "hms" or "dms"'


@register.filter
def target_extra_field(target, name):
    """
    Returns a ``TargetExtra`` value of the given name, if one exists.
    """
    try:
        return TargetExtra.objects.get(target=target, key=name).value
    except TargetExtra.DoesNotExist:
        return None


@register.inclusion_tag('tom_targets/partials/targetlist_select.html')
def select_target_js():
    """
    """
    return


@register.inclusion_tag('tom_targets/partials/aladin.html')
def aladin(target):
    """
    Displays Aladin skyview of the given target along with basic finder chart annotations including a compass
    and a scale bar. The resulting image is downloadable. This templatetag only works for sidereal targets.
    """
    return {'target': target}


@register.inclusion_tag('tom_targets/partials/target_table.html')
def target_table(targets):
    """
    Returns a partial for a table of targets, used in the target_list.html template
    by default
    """
    return {'targets': targets}



# Generated by Django 3.1 on 2020-08-18 20:34

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('tom_observations', '0008_observationgroup_cadence_parameters'),
    ]

    operations = [
        migrations.AddField(
            model_name='observationrecord',
            name='user',
            field=models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.DO_NOTHING, to=settings.AUTH_USER_MODEL),
        ),
    ]



# Generated by Django 2.2.2 on 2019-07-01 22:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_alerts', '0002_auto_20190626_2135'),
    ]

    operations = [
        migrations.AlterField(
            model_name='brokerquery',
            name='modified',
            field=models.DateTimeField(auto_now=True),
        ),
    ]



# Generated by Django 2.1.4 on 2018-12-06 21:57

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Target',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', models.CharField(help_text='The identifier for this object, e.g. Kelt-16b.', max_length=100, verbose_name='Identifier')),
                ('name', models.CharField(default='', help_text="The name of this target e.g. Barnard's star.", max_length=100, verbose_name='Name')),
                ('type', models.CharField(choices=[('SIDEREAL', 'Sidereal'), ('NON_SIDEREAL', 'Non-sidereal')], help_text='The type of this target.', max_length=100, verbose_name='Target Type')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='The time which this target was created in the TOM database.', verbose_name='Time Created')),
                ('modified', models.DateTimeField(auto_now=True, help_text='The time which this target was changed in the TOM database.', verbose_name='Last Modified')),
                ('ra', models.FloatField(blank=True, help_text='Right Ascension, in degrees.', null=True, verbose_name='Right Ascension')),
                ('dec', models.FloatField(blank=True, help_text='Declination, in degrees.', null=True, verbose_name='Declination')),
                ('epoch', models.FloatField(blank=True, help_text='Julian Years. Max 2100.', null=True, verbose_name='Epoch of Elements')),
                ('parallax', models.FloatField(blank=True, help_text='Parallax, in milliarcseconds.', null=True, verbose_name='Parallax')),
                ('pm_ra', models.FloatField(blank=True, help_text='Proper Motion: RA. Milliarsec/year.', null=True, verbose_name='Proper Motion (RA)')),
                ('pm_dec', models.FloatField(blank=True, help_text='Proper Motion: Dec. Milliarsec/year.', null=True, verbose_name='Proper Motion (Declination)')),
                ('galactic_lng', models.FloatField(blank=True, help_text='Galactic Longitude in degrees.', null=True, verbose_name='Galactic Longitude')),
                ('galactic_lat', models.FloatField(blank=True, help_text='Galactic Latitude in degrees.', null=True, verbose_name='Galactic Latitude')),
                ('distance', models.FloatField(blank=True, help_text='Parsecs.', null=True, verbose_name='Distance')),
                ('distance_err', models.FloatField(blank=True, help_text='Parsecs.', null=True, verbose_name='Distance Error')),
                ('mean_anomaly', models.FloatField(blank=True, help_text='Angle in degrees.', null=True, verbose_name='Mean Anomaly')),
                ('arg_of_perihelion', models.FloatField(blank=True, help_text='Argument of Perhihelion. J2000. Degrees.', null=True, verbose_name='Argument of Perihelion')),
                ('eccentricity', models.FloatField(blank=True, help_text='Eccentricity', null=True, verbose_name='Eccentricity')),
                ('lng_asc_node', models.FloatField(blank=True, help_text='Longitude of Ascending Node. J2000. Degrees.', null=True, verbose_name='Longitude of Ascending Node')),
                ('inclination', models.FloatField(blank=True, help_text='Inclination to the ecliptic. J2000. Degrees.', null=True, verbose_name='Inclination to the ecliptic')),
                ('mean_daily_motion', models.FloatField(blank=True, help_text='Degrees per day.', null=True, verbose_name='Mean Daily Motion')),
                ('semimajor_axis', models.FloatField(blank=True, help_text='In AU', null=True, verbose_name='Semimajor Axis')),
                ('ephemeris_period', models.FloatField(blank=True, help_text='Days', null=True, verbose_name='Ephemeris Period')),
                ('ephemeris_period_err', models.FloatField(blank=True, help_text='Days', null=True, verbose_name='Ephemeris Period Error')),
                ('ephemeris_epoch', models.FloatField(blank=True, help_text='Days', null=True, verbose_name='Ephemeris Epoch')),
                ('ephemeris_epoch_err', models.FloatField(blank=True, help_text='Days', null=True, verbose_name='Ephemeris Epoch Error')),
            ],
            options={
                'ordering': ('id',),
            },
        ),
        migrations.CreateModel(
            name='TargetExtra',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=200)),
                ('value', models.TextField()),
                ('target', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tom_targets.Target')),
            ],
        ),
        migrations.CreateModel(
            name='TargetList',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The name of the target list.', max_length=200)),
                ('created', models.DateTimeField(auto_now_add=True, help_text='The time which this target list was created in the TOM database.')),
                ('targets', models.ManyToManyField(to='tom_targets.Target')),
            ],
        ),
    ]



from io import StringIO
from urllib.parse import urlencode

from crispy_forms.bootstrap import FormActions
from crispy_forms.layout import HTML, Layout, Submit
from django import forms
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.management import call_command
from django_filters import (CharFilter, ChoiceFilter, DateTimeFromToRangeFilter, FilterSet, ModelMultipleChoiceFilter,
                            OrderingFilter)
from django_filters.views import FilterView
from django.shortcuts import redirect
from django.urls import reverse, reverse_lazy
from django.utils.safestring import mark_safe
from django.views.generic import View, TemplateView
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, DeleteView, FormView, UpdateView
from django.views.generic.list import ListView
from guardian.shortcuts import get_objects_for_user, assign_perm
from guardian.mixins import PermissionListMixin

from tom_common.hints import add_hint
from tom_common.mixins import Raise403PermissionRequiredMixin
from tom_dataproducts.forms import AddProductToGroupForm, DataProductUploadForm
from tom_observations.cadence import CadenceForm, get_cadence_strategy
from tom_observations.facility import get_service_class, get_service_classes
from tom_observations.facility import BaseManualObservationFacility
from tom_observations.forms import AddExistingObservationForm
from tom_observations.models import ObservationRecord, ObservationGroup, ObservationTemplate, DynamicCadence
from tom_targets.models import Target


class ObservationFilter(FilterSet):
    """
    Defines the available fields for filtering the list of ``ObservationRecord`` objects.
    """
    ordering = OrderingFilter(
        fields=['scheduled_start', 'scheduled_end', 'status', 'created', 'modified']
    )
    scheduled_start = DateTimeFromToRangeFilter()
    scheduled_end = DateTimeFromToRangeFilter
    observationgroup = ModelMultipleChoiceFilter(
        label='Observation Groups', queryset=ObservationGroup.objects.all()
    )

    class Meta:
        model = ObservationRecord
        fields = ['ordering', 'observation_id', 'target_id', 'observationgroup', 'facility', 'status']


class ObservationListView(FilterView):
    """
    View that displays all ``ObservationRecord`` objects.
    """
    filterset_class = ObservationFilter
    template_name = 'tom_observations/observation_list.html'
    paginate_by = 25
    model = ObservationRecord
    strict = False

    def get_queryset(self, *args, **kwargs):
        """
        Gets the set of ``ObservationRecord`` objects associated with the targets that the user is authorized to view.

        :returns: set of ObservationRecords
        :rtype: QuerySet
        """
        if settings.TARGET_PERMISSIONS_ONLY:
            return ObservationRecord.objects.filter(
                target__in=get_objects_for_user(self.request.user, 'tom_targets.view_target')
            )
        else:
            return get_objects_for_user(self.request.user, 'tom_observations.view_observationrecord')

    def get(self, request, *args, **kwargs):
        """
        Handles the GET requests to this view. If update_status is passed in the query parameters, calls the
        updatestatus management command to query for new statuses for ``ObservationRecord`` objects.

        :param request: request object for this GET request
        :type request: HTTPRequest
        """
        # QueryDict is immutable, and we want to append the remaining parameters to the redirect URL
        query_params = request.GET.copy()
        update_status = query_params.pop('update_status', False)
        if update_status:
            if not request.user.is_authenticated:
                return redirect(reverse('login'))
            out = StringIO()
            call_command('updatestatus', stdout=out)
            messages.info(request, out.getvalue())
            add_hint(request, mark_safe(
                              'Did you know updating observation statuses can be automated? Learn how in '
                              '<a href=https://tom-toolkit.readthedocs.io/en/stable/customization/automation.html>'
                              'the docs.</a>'))
            return redirect(f'{reverse("tom_observations:list")}?{urlencode(query_params)}')

        selected = request.GET.getlist('selected')
        observationgroups = request.GET.getlist('observationgroup')
        action = request.GET.get('action')
        if selected and observationgroups and action:
            observation_records = ObservationRecord.objects.filter(id__in=selected)
            groups = ObservationGroup.objects.filter(id__in=observationgroups)
            for group in groups:
                if action == 'add':
                    group.observation_records.add(*observation_records)
                if action == 'remove':
                    group.observation_records.remove(*observation_records)
                group.save()
            return redirect(reverse('tom_observations:list'))
        return super().get(request, *args, **kwargs)


# TODO: Ensure this template includes the ApplyObservationTemplate form at the top
class ObservationCreateView(LoginRequiredMixin, FormView):
    """
    View for creation/submission of an observation. Requires authentication.
    """
    template_name = 'tom_observations/observation_form.html'

    def get_target_id(self):
        """
        Parses the target id for the given observation from the query parameters.

        :returns: id of the target for observing
        :rtype: int
        """
        if self.request.method == 'GET':
            return self.request.GET.get('target_id')
        elif self.request.method == 'POST':
            return self.request.POST.get('target_id')

    def get_target(self):
        """
        Gets the target for observing from the database

        :returns: target for observing
        :rtype: Target
        """
        return Target.objects.get(pk=self.get_target_id())

    def get_facility(self):
        """
        Gets the facility from which the target is being observed from the query parameters

        :returns: facility name
        :rtype: str
        """
        return self.kwargs['facility']

    def get_facility_class(self):
        """
        Gets the facility interface class

        :returns: facility class name
        :rtype: str
        """
        return get_service_class(self.get_facility())

    def get_cadence_strategy_form(self):
        cadence_strategy = self.request.GET.get('cadence_strategy')
        if not cadence_strategy:
            return CadenceForm
        return get_cadence_strategy(cadence_strategy).form

    def get_context_data(self, **kwargs):
        """
        Adds the available observation types for the observing facility to the context object.

        :returns: context dictionary
        :rtype: dict
        """
        context = super(ObservationCreateView, self).get_context_data(**kwargs)

        # Populate initial values for each form and add them to the context. If the page
        # reloaded due to form errors, only repopulate the form that was submitted.
        observation_type_choices = []
        initial = self.get_initial()
        for observation_type, observation_form_class in self.get_facility_class().observation_forms.items():
            form_data = {**initial, **{'observation_type': observation_type}}
            # Repopulate the appropriate form with form data if the original submission was invalid
            if observation_type == self.request.POST.get('observation_type'):
                form_data.update(**self.request.POST.dict())
            observation_form_class = type(f'Composite{observation_type}Form',
                                          (self.get_cadence_strategy_form(), observation_form_class), {})
            observation_type_choices.append((observation_type, observation_form_class(initial=form_data)))
        context['observation_type_choices'] = observation_type_choices

        # Ensure correct tab is active if submission is unsuccessful
        context['active'] = self.request.POST.get('observation_type')

        target = Target.objects.get(pk=self.get_target_id())
        context['target'] = target
        return context

    def get_form_class(self):
        """
        Gets the observation form class for the facility and selected observation type in the query parameters.

        :returns: observation form
        :rtype: subclass of GenericObservationForm
        """
        observation_type = None
        if self.request.method == 'GET':
            observation_type = self.request.GET.get('observation_type')
        elif self.request.method == 'POST':
            observation_type = self.request.POST.get('observation_type')
        form_class = type(f'Composite{observation_type}Form',
                          (self.get_facility_class()().get_form(observation_type), self.get_cadence_strategy_form()),
                          {})
        return form_class

    def get_form(self):
        """
        Gets an instance of the form appropriate for the request.

        :returns: observation form
        :rtype: subclass of GenericObservationForm
        """
        form = super().get_form()
        if not settings.TARGET_PERMISSIONS_ONLY:
            form.fields['groups'].queryset = self.request.user.groups.all()
        form.helper.form_action = reverse(
            'tom_observations:create', kwargs=self.kwargs
        )
        return form

    def get_initial(self):
        """
        Populates the observation form with initial data including the id of the target to be observed, the facility at
        which the observation will take place, and the observation type desired.

        :returns: initial form data
        :rtype: dict
        """
        initial = super().get_initial()
        if not self.get_target_id():
            raise Exception('Must provide target_id')
        initial['target_id'] = self.get_target_id()
        initial['facility'] = self.get_facility()
        initial.update(self.request.GET.dict())
        return initial

    def post(self, request, *args, **kwargs):
        form = self.get_form()
        if form.is_valid():
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    def form_valid(self, form):
        """
        Runs after form validation. Submits the observation to the desired facility and creates an associated
        ``ObservationRecord``, then redirects to the detail page of the target to be observed.

        If the facility returns more than one record, a group is created and all observation
        records from the request are added to it.

        :param form: form containing observating request parameters
        :type form: subclass of GenericObservationForm
        """
        # Submit the observation
        facility = self.get_facility_class()
        target = self.get_target()
        observation_ids = facility().submit_observation(form.observation_payload())
        records = []

        for observation_id in observation_ids:
            # Create Observation record
            record = ObservationRecord.objects.create(
                target=target,
                user=self.request.user,
                facility=facility.name,
                parameters=form.serialize_parameters(),
                observation_id=observation_id
            )
            records.append(record)

        # TODO: redirect to observation list for multiple observations, observation detail otherwise

        if len(records) > 1 or form.cleaned_data.get('cadence_strategy'):
            observation_group = ObservationGroup.objects.create(name=form.cleaned_data['name'])
            observation_group.observation_records.add(*records)
            assign_perm('tom_observations.view_observationgroup', self.request.user, observation_group)
            assign_perm('tom_observations.change_observationgroup', self.request.user, observation_group)
            assign_perm('tom_observations.delete_observationgroup', self.request.user, observation_group)

            # TODO: Add a test case that includes a dynamic cadence submission
            if form.cleaned_data.get('cadence_strategy'):
                cadence_parameters = {}
                cadence_form = get_cadence_strategy(form.cleaned_data.get('cadence_strategy')).form
                for field in cadence_form().cadence_fields:
                    cadence_parameters[field] = form.cleaned_data.get(field)
                DynamicCadence.objects.create(
                    observation_group=observation_group,
                    cadence_strategy=form.cleaned_data.get('cadence_strategy'),
                    cadence_parameters=cadence_parameters,
                    active=True
                )

        if not settings.TARGET_PERMISSIONS_ONLY:
            groups = form.cleaned_data['groups']
            for record in records:
                assign_perm('tom_observations.view_observationrecord', groups, record)
                assign_perm('tom_observations.change_observationrecord', groups, record)
                assign_perm('tom_observations.delete_observationrecord', groups, record)

        return redirect(
            reverse('tom_targets:detail', kwargs={'pk': target.id})
        )


class ObservationRecordUpdateView(LoginRequiredMixin, UpdateView):
    """
    This view allows for the updating of the observation id, which will eventually be expanded to more fields.
    """
    model = ObservationRecord
    fields = ['observation_id']
    template_name = 'tom_observations/observationupdate_form.html'

    def get_success_url(self):
        return reverse('tom_observations:detail', kwargs={'pk': self.get_object().id})


class ObservationRecordCancelView(LoginRequiredMixin, View):

    def get(self, request, *args, **kwargs):
        obsr_id = self.kwargs.get('pk')
        obsr = ObservationRecord.objects.get(id=obsr_id)
        facility = get_service_class(obsr.facility)()
        try:
            success = facility.cancel_observation(obsr.observation_id)
            if success:
                messages.success(self.request, f'Successfully cancelled observation {obsr}')
                facility.update_observation_status(obsr.observation_id)
            else:
                messages.error(self.request, 'Unable to cancel observation.')
        except forms.ValidationError as ve:
            messages.error(self.request, f'Unable to cancel observation: {ve}')

        return redirect(reverse('tom_observations:detail', kwargs={'pk': obsr.id}))


class AddExistingObservationView(LoginRequiredMixin, FormView):
    """
    View for associating a pre-existing observation with a target. Requires authentication.

    The GET view returns a confirmation page for adding duplicate ObservationRecords. Two duplicates are any two
    ObservationRecords with the same target_id, facility, and observation_id.

    The POST view validates the form and redirects to the confirmation page if the confirm flag isn't set.

    This view is intended to be navigated to via the existing_observation_button templatetag, as the
    AddExistingObservationForm has a hidden confirmation checkbox selected by default.
    """
    template_name = 'tom_observations/existing_observation_confirm.html'
    form_class = AddExistingObservationForm

    def get_form(self):
        form = super().get_form()
        form.fields['facility'].widget = forms.HiddenInput()
        form.fields['observation_id'].widget = forms.HiddenInput()
        if self.request.method == 'GET':
            target_id = self.request.GET.get('target_id')
        elif self.request.method == 'POST':
            target_id = self.request.POST.get('target_id')
        cancel_url = reverse('home')
        if target_id:
            cancel_url = reverse('tom_targets:detail', kwargs={'pk': target_id})
        form.helper.layout = Layout(
            HTML('''<p>An observation record already exists in your TOM for this combination of observation ID,
                 facility, and target. Are you sure you want to create this record?</p>'''),
            'target_id',
            'facility',
            'observation_id',
            'confirm',
            FormActions(
                Submit('confirm', 'Confirm'),
                HTML(f'<a class="btn btn-outline-primary" href={cancel_url}>Cancel</a>')
            )
        )
        return form

    def get_initial(self):
        """
        Populates the ``ManualObservationForm`` hidden field for target id with the id from the specified target.

        :returns: initial form data
        :rtype: dict
        """
        if self.request.method == 'GET':
            params = self.request.GET.dict()
            params['confirm'] = True
            return params

    def form_valid(self, form):
        """
        Runs after form validation. Creates a new ``ObservationRecord`` associated with the specified target and
        facility.
        """
        records = ObservationRecord.objects.filter(target_id=form.cleaned_data['target_id'],
                                                   facility=form.cleaned_data['facility'],
                                                   observation_id=form.cleaned_data['observation_id'])

        if records and not form.cleaned_data.get('confirm'):
            return redirect(reverse('tom_observations:add-existing') + '?' + self.request.POST.urlencode())
        else:
            ObservationRecord.objects.create(
                target_id=form.cleaned_data['target_id'],
                facility=form.cleaned_data['facility'],
                parameters={},
                observation_id=form.cleaned_data['observation_id']
            )
            observation_id = form.cleaned_data['observation_id']
            messages.success(self.request, f'Successfully associated observation record {observation_id}')
        return redirect(reverse(
            'tom_targets:detail', kwargs={'pk': form.cleaned_data['target_id']})
        )


class ObservationRecordDetailView(DetailView):
    """
    View for displaying the details of an ``ObservationRecord`` object.
    """
    model = ObservationRecord

    def get_queryset(self, *args, **kwargs):
        """
        Gets the set of ``ObservationRecord`` objects associated with targets that the current user is authorized to
        view.

        :returns: set of ObservationRecords
        :rtype: QuerySet
        """
        if settings.TARGET_PERMISSIONS_ONLY:
            return ObservationRecord.objects.filter(
                target__in=get_objects_for_user(self.request.user, 'tom_targets.view_target')
            )
        else:
            return get_objects_for_user(self.request.user, 'tom_observations.view_observationrecord')

    def get_context_data(self, *args, **kwargs):
        """
        Adds a number of items to the context object for this view, including the form for adding resulting
        ``DataProduct`` objects to a ``DataProductGroup``, the ``DataProduct`` objects associated with the
        ``ObservationRecord``, and the most recent image from this ``ObservationRecord``. It also populates the
        ``DataProductUploadForm`` hidden fields with initial data.

        :returns: context dictionary
        :rtype: dict
        """
        context = super().get_context_data(*args, **kwargs)
        context['form'] = AddProductToGroupForm()
        service_class = get_service_class(self.object.facility)
        context['editable'] = isinstance(service_class(), BaseManualObservationFacility)
        context['data_products'] = service_class().all_data_products(self.object)
        context['can_be_cancelled'] = self.object.status not in service_class().get_terminal_observing_states()
        newest_image = None
        for data_product in context['data_products']['saved']:
            newest_image = data_product if (not newest_image or data_product.modified > newest_image.modified) and \
                data_product.get_file_extension() == '.fits' else newest_image
        if newest_image:
            context['image'] = newest_image.get_image_data()
        data_product_upload_form = DataProductUploadForm(
            initial={
                'observation_record': self.get_object(),
                'referrer': reverse('tom_observations:detail', args=(self.get_object().id,))
            }
        )
        context['data_product_form'] = data_product_upload_form
        return context


class ObservationGroupCreateView(LoginRequiredMixin, CreateView):
    """
    View that handles the creation of ``ObservationGroup`` objects. Requires authentication.
    """
    model = ObservationGroup
    fields = ['name']
    success_url = reverse_lazy('tom_observations:group-list')

    def form_valid(self, form):
        """
        Runs after form validation. Saves the observation group and assigns the user's permissions to the group.

        :param form: Form data for observation group creation
        :type form: django.forms.ModelForm
        """
        obj = form.save(commit=False)
        obj.save()
        assign_perm('tom_observations.view_observationgroup', self.request.user, obj)
        assign_perm('tom_observations.change_observationgroup', self.request.user, obj)
        assign_perm('tom_observations.delete_observationgroup', self.request.user, obj)
        return super().form_valid(form)


class ObservationGroupListView(PermissionListMixin, ListView):
    """
    View that handles the display of ``ObservationGroup``.
    Requires authorization.
    """
    permission_required = 'tom_observations.view_observationgroup'
    model = ObservationGroup
    paginate_by = 25


class ObservationGroupDeleteView(Raise403PermissionRequiredMixin, DeleteView):
    """
    View that handles the deletion of ``ObservationGroup`` objects. Requires authorization.
    """
    permission_required = 'tom_observations.delete_observationgroup'
    model = ObservationGroup
    success_url = reverse_lazy('tom_observations:group-list')


class ObservationTemplateFilter(FilterSet):
    """
    Defines the available fields for filtering the list of ``ObservationTemplate`` objects.
    """
    facility = ChoiceFilter(
        choices=[(k, k) for k in get_service_classes().keys()]
    )
    name = CharFilter(lookup_expr='icontains')

    class Meta:
        model = ObservationTemplate
        fields = ['name', 'facility']


class ObservationTemplateListView(FilterView):
    """
    Displays the observing strategies that exist in the TOM.
    """
    model = ObservationTemplate
    filterset_class = ObservationTemplateFilter
    template_name = 'tom_observations/observationtemplate_list.html'

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context['installed_facilities'] = get_service_classes()
        return context


class ObservationTemplateCreateView(FormView):
    """
    Displays the form for creating a new observation template. Uses the observation template form specified in the
    respective facility class.
    """
    template_name = 'tom_observations/observationtemplate_form.html'

    def get_facility_name(self):
        return self.kwargs['facility']

    def get_form_class(self):
        facility_name = self.get_facility_name()

        if not facility_name:
            raise ValueError('Must provide a facility name')

        # TODO: modify this to work with all LCO forms
        return get_service_class(facility_name)().get_template_form(None)

    def get_form(self, form_class=None):
        form = super().get_form()
        form.helper.form_action = reverse('tom_observations:template-create',
                                          kwargs={'facility': self.get_facility_name()})
        return form

    def get_initial(self):
        initial = super().get_initial()
        initial['facility'] = self.get_facility_name()
        initial.update(self.request.GET.dict())
        return initial

    def form_valid(self, form):
        form.save()
        return redirect(reverse('tom_observations:template-list'))


class ObservationTemplateUpdateView(LoginRequiredMixin, FormView):
    """
    View for updating an existing observation template.
    """
    template_name = 'tom_observations/observationtemplate_form.html'

    def get_object(self):
        return ObservationTemplate.objects.get(pk=self.kwargs['pk'])

    def get_form_class(self):
        self.object = self.get_object()
        return get_service_class(self.object.facility)().get_template_form(None)

    def get_form(self):
        form = super().get_form()
        form.helper.form_action = reverse(
            'tom_observations:template-update', kwargs={'pk': self.object.id}
        )
        return form

    def get_initial(self):
        initial = super().get_initial()
        initial.update(self.object.parameters)
        initial['facility'] = self.object.facility
        return initial

    def form_valid(self, form):
        form.save(template_id=self.object.id)
        return redirect(reverse('tom_observations:template-list'))


class ObservationTemplateDeleteView(LoginRequiredMixin, DeleteView):
    """
    Deletes an observation template.
    """
    model = ObservationTemplate
    success_url = reverse_lazy('tom_observations:template-list')


class FacilityStatusView(TemplateView):
    template_name = 'tom_observations/facility_status.html'



from datetime import datetime
from dateutil.parser import parse

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.forms.models import model_to_dict
from django.urls import reverse

from tom_common.hooks import run_hook

GLOBAL_TARGET_FIELDS = ['name', 'type']

SIDEREAL_FIELDS = GLOBAL_TARGET_FIELDS + [
    'ra', 'dec', 'epoch', 'pm_ra', 'pm_dec', 'galactic_lng', 'galactic_lat', 'distance', 'distance_err'
]

NON_SIDEREAL_FIELDS = GLOBAL_TARGET_FIELDS + [
    'scheme', 'mean_anomaly', 'arg_of_perihelion', 'lng_asc_node', 'inclination', 'mean_daily_motion', 'semimajor_axis',
    'eccentricity', 'epoch_of_elements', 'epoch_of_perihelion', 'ephemeris_period', 'ephemeris_period_err',
    'ephemeris_epoch', 'ephemeris_epoch_err', 'perihdist'
]

REQUIRED_SIDEREAL_FIELDS = ['ra', 'dec']
REQUIRED_NON_SIDEREAL_FIELDS = [
    'scheme', 'epoch_of_elements', 'inclination', 'lng_asc_node', 'arg_of_perihelion', 'eccentricity',
]
# Additional non-sidereal fields that are required for specific orbital element
# schemes
REQUIRED_NON_SIDEREAL_FIELDS_PER_SCHEME = {
    'MPC_COMET': ['perihdist', 'epoch_of_perihelion'],
    'MPC_MINOR_PLANET': ['mean_anomaly', 'semimajor_axis'],
    'JPL_MAJOR_PLANET': ['mean_daily_motion', 'mean_anomaly', 'semimajor_axis']
}


class Target(models.Model):
    """
    Class representing a target in a TOM

    :param name: The name of this target e.g. Barnard\'s star.
    :type name: str

    :param type: The type of this target.
    :type type: str

    :param created: The time at which this target was created in the TOM database.
    :type type: datetime

    :param modified: The time at which this target was changed in the TOM database.
    :type type:

    :param ra: Right Ascension, in degrees.
    :type ra: float

    :param dec: Declination, in degrees.
    :type dec: float

    :param epoch: Julian Years. Max 2100.
    :type epoch: float

    :param parallax: Parallax, in milliarcseconds.
    :type parallax: float

    :param pm_ra: Proper Motion: RA. Milliarsec/year.
    :type pm_ra: float

    :param pm_dec: Proper Motion: Dec. Milliarsec/year.
    :type pm_dec: float

    :param galactic_lng: Galactic Longitude in degrees.
    :type galactic_lng: float

    :param galactic_lat: Galactic Latitude in degrees.
    :type galactic_lat: float

    :param distance: Parsecs.
    :type distance: float

    :param distance_err: Parsecs.
    :type distance_err: float

    :param scheme: Orbital Element Scheme
    :type scheme: str

    :param epoch_of_elements: Epoch of elements in JD.
    :type epoch_of_elements: float

    :param mean_anomaly: Angle in degrees.
    :type mean_anomaly: float

    :param arg_of_perihelion: Argument of Perhihelion. J2000. Degrees.
    :type arg_of_perihelion: float

    :param eccentricity: Eccentricity
    :type eccentricity: float

    :param lng_asc_node: Longitude of Ascending Node. J2000. Degrees.
    :type lng_asc_node: float

    :param inclination: Inclination to the ecliptic. J2000. Degrees.
    :type inclination: float

    :param mean_daily_motion: Degrees per day.
    :type mean_daily_motion: float

    :param semimajor_axis: Semimajor Axis in AU
    :type semimajor_axis: float

    :param epoch_of_perihelion: Julian Date.
    :type epoch_of_perihelion: float

    :param ephemeris_period: Ephemeris period in days
    :type ephemeris_period: float

    :param ephemeris_period_err: Days
    :type ephemeris_period_err: float

    :param ephemeris_epoch: Days
    :type ephemeris_epoch: float

    :param ephemeris_epoch_err: Days
    :type ephemeris_epoch_err: float
    """

    SIDEREAL = 'SIDEREAL'
    NON_SIDEREAL = 'NON_SIDEREAL'
    TARGET_TYPES = ((SIDEREAL, 'Sidereal'), (NON_SIDEREAL, 'Non-sidereal'))

    TARGET_SCHEMES = (
        ('MPC_MINOR_PLANET', 'MPC Minor Planet'),
        ('MPC_COMET', 'MPC Comet'),
        ('JPL_MAJOR_PLANET', 'JPL Major Planet')
    )

    name = models.CharField(
        max_length=100, default='', verbose_name='Name', help_text='The name of this target e.g. Barnard\'s star.',
        unique=True
    )
    type = models.CharField(
        max_length=100, choices=TARGET_TYPES, verbose_name='Target Type', help_text='The type of this target.'
    )
    created = models.DateTimeField(
        auto_now_add=True, verbose_name='Time Created',
        help_text='The time which this target was created in the TOM database.'
    )
    modified = models.DateTimeField(
        auto_now=True, verbose_name='Last Modified',
        help_text='The time which this target was changed in the TOM database.'
    )
    ra = models.FloatField(
        null=True, blank=True, verbose_name='Right Ascension', help_text='Right Ascension, in degrees.'
    )
    dec = models.FloatField(
        null=True, blank=True, verbose_name='Declination', help_text='Declination, in degrees.'
    )
    epoch = models.FloatField(
        null=True, blank=True, verbose_name='Epoch', help_text='Julian Years. Max 2100.'
    )
    parallax = models.FloatField(
        null=True, blank=True, verbose_name='Parallax', help_text='Parallax, in milliarcseconds.'
    )
    pm_ra = models.FloatField(
        null=True, blank=True, verbose_name='Proper Motion (RA)', help_text='Proper Motion: RA. Milliarsec/year.'
    )
    pm_dec = models.FloatField(
        null=True, blank=True, verbose_name='Proper Motion (Declination)',
        help_text='Proper Motion: Dec. Milliarsec/year.'
    )
    galactic_lng = models.FloatField(
        null=True, blank=True, verbose_name='Galactic Longitude', help_text='Galactic Longitude in degrees.'
    )
    galactic_lat = models.FloatField(
        null=True, blank=True, verbose_name='Galactic Latitude', help_text='Galactic Latitude in degrees.'
    )
    distance = models.FloatField(
        null=True, blank=True, verbose_name='Distance', help_text='Parsecs.'
    )
    distance_err = models.FloatField(
        null=True, blank=True, verbose_name='Distance Error', help_text='Parsecs.'
    )
    scheme = models.CharField(
        max_length=50, choices=TARGET_SCHEMES, verbose_name='Orbital Element Scheme', default='', blank=True
    )
    epoch_of_elements = models.FloatField(
        null=True, blank=True, verbose_name='Epoch of Elements', help_text='Julian date.'
    )
    mean_anomaly = models.FloatField(
        null=True, blank=True, verbose_name='Mean Anomaly', help_text='Angle in degrees.'
    )
    arg_of_perihelion = models.FloatField(
        null=True, blank=True, verbose_name='Argument of Perihelion',
        help_text='Argument of Perhihelion. J2000. Degrees.'
    )
    eccentricity = models.FloatField(
        null=True, blank=True, verbose_name='Eccentricity', help_text='Eccentricity'
    )
    lng_asc_node = models.FloatField(
        null=True, blank=True, verbose_name='Longitude of Ascending Node',
        help_text='Longitude of Ascending Node. J2000. Degrees.'
    )
    inclination = models.FloatField(
        null=True, blank=True, verbose_name='Inclination to the ecliptic',
        help_text='Inclination to the ecliptic. J2000. Degrees.'
    )
    mean_daily_motion = models.FloatField(
        null=True, blank=True, verbose_name='Mean Daily Motion', help_text='Degrees per day.'
    )
    semimajor_axis = models.FloatField(
        null=True, blank=True, verbose_name='Semimajor Axis', help_text='In AU'
    )
    epoch_of_perihelion = models.FloatField(
        null=True, blank=True, verbose_name='Epoch of Perihelion', help_text='Julian Date.'
    )
    ephemeris_period = models.FloatField(
        null=True, blank=True, verbose_name='Ephemeris Period', help_text='Days'
    )
    ephemeris_period_err = models.FloatField(
        null=True, blank=True, verbose_name='Ephemeris Period Error', help_text='Days'
    )
    ephemeris_epoch = models.FloatField(
        null=True, blank=True, verbose_name='Ephemeris Epoch', help_text='Days'
    )
    ephemeris_epoch_err = models.FloatField(
        null=True, blank=True, verbose_name='Ephemeris Epoch Error', help_text='Days'
    )
    perihdist = models.FloatField(
        null=True, blank=True, verbose_name='Perihelion Distance', help_text='AU'
    )

    @transaction.atomic
    def save(self, *args, **kwargs):
        """
        Saves Target model data to the database, including extra fields. After saving to the database, also runs the
        hook ``target_post_save``. The hook run is the one specified in ``settings.py``.

        :Keyword Arguments:
            * extras (`dict`): dictionary of key/value pairs representing target attributes
        """
        extras = kwargs.pop('extras', {})
        names = kwargs.pop('names', [])

        created = False if self.id else True
        super().save(*args, **kwargs)

        if created:
            for extra_field in settings.EXTRA_FIELDS:
                if extra_field.get('default') is not None:
                    TargetExtra(target=self, key=extra_field['name'], value=extra_field.get('default')).save()

        for k, v in extras.items():
            target_extra, _ = TargetExtra.objects.get_or_create(target=self, key=k)
            target_extra.value = v
            target_extra.save()

        for name in names:
            name, _ = TargetName.objects.get_or_create(target=self, name=name)
            name.save()

        if not created:
            run_hook('target_post_save', target=self, created=created)

    def validate_unique(self, *args, **kwargs):
        """
        Ensures that Target.name and all aliases of the target are unique. Called automatically on save.
        """
        super().validate_unique(*args, **kwargs)
        for alias in self.aliases.all():
            if alias.name == self.name:
                raise ValidationError('Target name and target aliases must be unique')

    def __str__(self):
        return str(self.name)

    def get_absolute_url(self):
        return reverse('targets:detail', kwargs={'pk': self.id})

    def featured_image(self):
        """
        Gets the ``DataProduct`` associated with this ``Target`` that is a FITS file and is uniquely marked as
        "featured".

        :returns: ``DataProduct`` with data_product_type of ``fits_file`` and featured as ``True``
        :rtype: DataProduct
        """
        return self.dataproduct_set.filter(data_product_type='fits_file', featured=True).first()

    @property
    def names(self):
        """
        Gets a list with the name and aliases of this target

        :returns: list of all names and `TargetName` values associated with this target
        :rtype: list
        """
        return [self.name] + [alias.name for alias in self.aliases.all()]

    @property
    def future_observations(self):
        """
        Gets all observations scheduled for this ``Target``

        :returns: List of ``ObservationRecord`` objects without a terminal status
        :rtype: list
        """
        return [
            obs for obs in self.observationrecord_set.exclude(status='').order_by('scheduled_start') if not obs.terminal
        ]

    @property
    def extra_fields(self):
        """
        Gets all ``TargetExtra`` fields associated with this ``Target``, provided the key is defined in ``settings.py``
        ``EXTRA_FIELDS``

        :returns: Dictionary of key/value pairs representing target attributes
        :rtype: dict
        """
        defined_extras = [extra_field['name'] for extra_field in settings.EXTRA_FIELDS]
        types = {extra_field['name']: extra_field['type'] for extra_field in settings.EXTRA_FIELDS}
        return {te.key: te.typed_value(types[te.key])
                for te in self.targetextra_set.filter(key__in=defined_extras)}

    @property
    def tags(self):
        """
        Gets all ``TargetExtra`` fields associated with this ``Target``, provided the key is `NOT` defined in
        ``settings.py`` ``EXTRA_FIELDS``

        :returns: Dictionary of key/value pairs representing target attributes
        :rtype: dict
        """
        defined_extras = [extra_field['name'] for extra_field in settings.EXTRA_FIELDS]
        return {te.key: te.value for te in self.targetextra_set.exclude(key__in=defined_extras)}

    def as_dict(self):
        """
        Returns dictionary representation of attributes, excluding all attributes not associated with the ``type`` of
        this ``Target``.

        :returns: Dictionary of key/value pairs representing target attributes
        :rtype: dict
        """
        if self.type == self.SIDEREAL:
            fields_for_type = SIDEREAL_FIELDS
        elif self.type == self.NON_SIDEREAL:
            fields_for_type = NON_SIDEREAL_FIELDS
        else:
            fields_for_type = GLOBAL_TARGET_FIELDS

        return model_to_dict(self, fields=fields_for_type)


class TargetName(models.Model):
    """
    Class representing an alternative name for a ``Target``.

    :param target: The ``Target`` object this ``TargetName`` is associated with.

    :param name: The name that this ``TargetName`` object represents.
    :type name: str

    :param created: The time at which this target name was created in the TOM database.
    :type created: datetime

    :param modified: The time at which this target name was modified in the TOM database.
    :type modified: datetime
    """
    target = models.ForeignKey(Target, on_delete=models.CASCADE, related_name='aliases')
    name = models.CharField(max_length=100, unique=True, verbose_name='Alias')
    created = models.DateTimeField(
        auto_now_add=True, help_text='The time which this target name was created.'
    )
    modified = models.DateTimeField(
        auto_now=True, verbose_name='Last Modified',
        help_text='The time which this target name was changed in the TOM database.'
    )

    def __str__(self):
        return self.name

    def validate_unique(self, *args, **kwargs):
        """
        Ensures that Target.name and all aliases of the target are unique. Called automatically on save.
        """
        super().validate_unique(*args, **kwargs)
        if self.name == self.target.name:
            raise ValidationError(f'''Alias {self.name} has a conflict with the primary name of the target
                                      {self.target.name} (id={self.target.id})''')


class TargetExtra(models.Model):
    """
    Class representing a list of targets in a TOM.

    :param target: The ``Target`` object this ``TargetExtra`` is associated with.

    :param key: Denotation of the value represented by this ``TargetExtra`` object.
    :type key: str

    :param value: Value of the field stored in this object.
    :type value: str

    :param float_value: Float representation of the ``value`` field for this object, if applicable.
    :type float_value: float

    :param bool_value: Boolean representation of the ``value`` field for this object, if applicable.
    :type bool_value: bool

    :param time_value: Datetime representation of the ``value`` field for this object, if applicable.
    :type time_value: datetime
    """
    target = models.ForeignKey(Target, on_delete=models.CASCADE)
    key = models.CharField(max_length=200)
    value = models.TextField(blank=True, default='')
    float_value = models.FloatField(null=True, blank=True)
    bool_value = models.BooleanField(null=True, blank=True)
    time_value = models.DateTimeField(null=True, blank=True)

    class Meta:
        unique_together = ['target', 'key']

    def __str__(self):
        return f'{self.key}: {self.value}'

    def save(self, *args, **kwargs):
        """
        Saves TargetExtra model data to the database. In the process, converts the string value of the ``TargetExtra``
        to the appropriate type, and stores it in the corresponding field as well.
        """
        try:
            self.float_value = float(self.value)
        except (TypeError, ValueError, OverflowError):
            self.float_value = None
        try:
            self.bool_value = bool(self.value)
        except (TypeError, ValueError, OverflowError):
            self.bool_value = None
        try:
            if isinstance(self.value, datetime):
                self.time_value = self.value
            else:
                self.time_value = parse(self.value)
        except (TypeError, ValueError, OverflowError):
            self.time_value = None

        super().save(*args, **kwargs)

    def typed_value(self, type_val):
        """
        Returns the value of this ``TargetExtra`` in the corresponding type provided by the caller. If the type is
        invalid, returns the string representation.

        :param type_val: Requested type of the ``TargetExtra`` ``value`` field
        :type type_val: str

        :returns: Requested typed value field of this object
        :rtype: float, boolean, datetime, or str
        """
        if type_val == 'number':
            return self.float_value
        if type_val == 'boolean':
            return self.bool_value
        if type_val == 'datetime':
            return self.time_value

        return self.value


class TargetList(models.Model):
    """
    Class representing a list of targets in a TOM.

    :param name: The name of the target list
    :type name: str

    :param targets: Set of ``Target`` objects associated with this ``TargetList``

    :param created: The time at which this target list was created.
    :type created: datetime

    :param modified: The time at which this target list was modified in the TOM database.
    :type modified: datetime
    """
    name = models.CharField(max_length=200, help_text='The name of the target list.')
    targets = models.ManyToManyField(Target)
    created = models.DateTimeField(
        auto_now_add=True, help_text='The time which this target list was created in the TOM database.'
    )
    modified = models.DateTimeField(
        auto_now=True, verbose_name='Last Modified',
        help_text='The time which this target list was changed in the TOM database.'
    )

    class Meta:
        ordering = ('-created', 'name',)

    def __str__(self):
        return self.name



from copy import deepcopy
from unittest.mock import patch

from django.contrib.auth.models import Group, User
from django.test import override_settings
from django.urls import reverse
from django.utils import timezone
from guardian.shortcuts import assign_perm
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.test import APITestCase

from tom_observations.api_views import ObservationRecordViewSet
from tom_observations.models import ObservationGroup
from tom_observations.tests.factories import ObservingRecordFactory
from tom_targets.tests.factories import SiderealTargetFactory


@override_settings(TOM_FACILITY_CLASSES=['tom_observations.tests.utils.FakeRoboticFacility',
                                         'tom_observations.tests.utils.FakeManualFacility'],
                   TARGET_PERMISSIONS_ONLY=True)
class TestObservationViewset(APITestCase):
    def setUp(self):
        self.user = User.objects.create(username='testuser')
        self.st = SiderealTargetFactory.create(name='testtarget')
        self.st2 = SiderealTargetFactory.create(name='testtarget2')
        self.st3 = SiderealTargetFactory.create(name='testtarget3')

        self.obsr = ObservingRecordFactory.create(target_id=self.st.id, user=self.user)
        self.obsr2 = ObservingRecordFactory.create(target_id=self.st2.id)
        self.obsr3 = ObservingRecordFactory.create(target_id=self.st3.id)

        assign_perm('tom_targets.view_target', self.user, self.st2)

        self.client.force_login(self.user)

    def test_observation_detail(self):
        """Test observation API detail endpoint."""
        with self.subTest('Test that a user can view an ObservationRecord for an ObservationRecord they submitted.'):
            response = self.client.get(reverse('api:observations-detail', args=(self.obsr.id,)))
            self.assertEqual(response.json()['id'], self.obsr.id)

        with self.subTest('Test that a user can view an ObservationRecord for a target they can view.'):
            response = self.client.get(reverse('api:observations-detail', args=(self.obsr2.id,)))
            self.assertEqual(response.json()['id'], self.obsr2.id)

        with self.subTest('Test that a user cannot view an ObservationRecord for a target they cannot view.'):
            response = self.client.get(reverse('api:observations-detail', args=(self.obsr3.id,)))
            self.assertContains(response, 'Not found.', status_code=status.HTTP_404_NOT_FOUND)

    def test_observation_list(self):
        """Test observation API list endpoint."""
        response = self.client.get(reverse('api:observations-list'))
        self.assertEqual(response.json()['count'], 2)
        self.assertContains(response, f'"id":{self.obsr.id}')
        self.assertContains(response, f'"id":{self.obsr2.id}')
        self.assertNotContains(response, f'"id":{self.obsr3.id}')

    def test_observation_submit(self):
        """Test observation API submit endpoint."""
        form_data = {
            'target_id': self.st.id,
            'facility': 'FakeRoboticFacility',
            'observation_type': 'OBSERVATION',
            'observing_parameters': {
                'test_input': 'gnomes'
            }
        }
        response = self.client.post(reverse('api:observations-list'), data=form_data, follow=True)
        self.assertContains(response, 'fakeid', status_code=status.HTTP_201_CREATED)
        self.assertDictContainsSubset({'test_input': 'gnomes'}, response.json()[0]['parameters'])

    def test_observation_submit_invalid_parameters(self):
        """Test observation API submit endpoint with unsuccessful submissions."""
        form_data = {
            'target_id': self.st.id,
            'facility': 'FakeRoboticFacility',
            'observation_type': 'OBSERVATION',
            'observing_parameters': {
                'test_input': 'gnomes'
            }
        }

        required_fields = ['facility', 'observation_type', 'target_id', 'observing_parameters']
        for field_name in required_fields:
            with self.subTest('Test that all required fields return 400 if absent.'):
                missing_form_data = deepcopy(form_data)
                missing_form_data.pop(field_name)
                response = self.client.post(reverse('api:observations-list'), data=missing_form_data, follow=True)
                self.assertContains(response,
                                    f'Missing required field \'{field_name}\'.',
                                    status_code=status.HTTP_400_BAD_REQUEST)

        with self.subTest('Test that a unexpected exception returns 400.'):
            with patch('tom_observations.api_views.get_service_class') as mock_get_service_class:
                mock_get_service_class.side_effect = Exception('An error occurred.')
                response = self.client.post(reverse('api:observations-list'), data=form_data, follow=True)
                self.assertContains(response, 'An error occurred.', status_code=status.HTTP_400_BAD_REQUEST)

        with self.subTest('Test that an invalid observation form returns 400.'):
            bad_form_data = deepcopy(form_data)
            bad_form_data['observing_parameters'].pop('test_input')  # test_input is supposed to take a string
            response = self.client.post(reverse('api:observations-list'), data=bad_form_data, follow=True)
            self.assertContains(response,
                                'This field is required.',
                                status_code=status.HTTP_400_BAD_REQUEST)

    def test_observation_submit_cadence(self):
        """Test observation API submit endpoint with cadences."""
        form_data = {
            'name': 'Test Cadence',
            'target_id': self.st.id,
            'facility': 'FakeRoboticFacility',
            'observation_type': 'OBSERVATION',
            'observing_parameters': {
                'test_input': 'gnomes',
            },
            'cadence': {
                'cadence_strategy': 'ResumeCadenceAfterFailureStrategy',
                'cadence_frequency': 24,
            }
        }

        response = self.client.post(reverse('api:observations-list'), data=form_data, follow=True)
        self.assertContains(response, 'fakeid', status_code=status.HTTP_201_CREATED)
        self.assertDictContainsSubset({'name': f'{self.st.name} at FakeRoboticFacility'},
                                      response.json()[0].get('observation_groups', [])[0])
        self.assertIn(
            'ResumeCadenceAfterFailureStrategy with parameters {\'cadence_frequency\': 24}',
            response.json()[0].get('observation_groups', [])[0].get('dynamic_cadences', []))

    @patch('tom_observations.tests.utils.FakeRoboticFacility.submit_observation')
    def test_observation_multiple_records(self, mock_submit_observation):
        """Test observation API submit endpoint with multiple returned records."""
        mock_submit_observation.return_value = ['fakeid1', 'fakeid2']

        form_data = {
            'name': 'Test Multiple Records',
            'target_id': self.st.id,
            'facility': 'FakeRoboticFacility',
            'observation_type': 'OBSERVATION',
            'observing_parameters': {
                'test_input': 'gnomes',
            },
        }

        response = self.client.post(reverse('api:observations-list'), data=form_data, follow=True)
        self.assertContains(response, 'fakeid', status_code=status.HTTP_201_CREATED)
        self.assertDictContainsSubset({'name': f'{self.st.name} at FakeRoboticFacility'},
                                      response.json()[0].get('observation_groups', [])[0])

    def test_observation_submit_cadence_invalid_parameters(self):
        """Test observation API submit endpoint with cadences that are unsuccessful submissions."""
        form_data = {
            'name': 'Test Cadence',
            'target_id': self.st.id,
            'facility': 'FakeRoboticFacility',
            'observation_type': 'OBSERVATION',
            'observing_parameters': {
                'test_input': 'gnomes',
            },
            'cadence': {
                'cadence_strategy': 'ResumeCadenceAfterFailureStrategy',
                'cadence_frequency': 24,
            }
        }

        with self.subTest('Test that a missing cadence strategy returns a 400.'):
            bad_form_data = deepcopy(form_data)
            bad_form_data['cadence'].pop('cadence_strategy')
            response = self.client.post(reverse('api:observations-list'), data=bad_form_data, follow=True)
            self.assertContains(response,
                                'cadence_strategy must be included to initiate a DynamicCadence.',
                                status_code=status.HTTP_400_BAD_REQUEST)

        with self.subTest('Test that a missing cadence strategy returns a 400.'):
            bad_form_data = deepcopy(form_data)
            bad_form_data['cadence'].pop('cadence_frequency')
            response = self.client.post(reverse('api:observations-list'), data=bad_form_data, follow=True)
            self.assertContains(response,
                                'This field is required.',
                                status_code=status.HTTP_400_BAD_REQUEST)
            self.assertTrue(len(ObservationGroup.objects.all()), 0)

        with self.subTest('Test that a serializer failure returns a 400 and deletes objects.'):
            with patch.object(ObservationRecordViewSet, 'perform_create') as mock_serializer:
                mock_serializer.side_effect = ValidationError('test')
                response = self.client.post(reverse('api:observations-list'), data=form_data, follow=True)
                self.assertContains(response,
                                    'Observation submission successful, but failed to create a corresponding',
                                    status_code=status.HTTP_400_BAD_REQUEST)

    @patch('tom_observations.tests.utils.FakeRoboticFacility.get_observation_status')
    def test_observation_cancel(self, mock_get_status):
        mock_get_status.return_value = {'state': 'CANCELED',
                                        'scheduled_start': timezone.now(),
                                        'scheduled_end': timezone.now()}
        user2 = User.objects.create(username='testuser2')
        self.obsr2.facility = 'FakeRoboticFacility'
        self.obsr2.save()

        response = self.client.patch(reverse('api:observations-cancel', kwargs={'pk': self.obsr2.id}))
        self.assertContains(response, 'CANCELED')

        self.client.force_login(user2)
        response = self.client.patch(reverse('api:observations-cancel', kwargs={'pk': self.obsr2.id}))
        self.assertContains(response, 'Not found.', status_code=status.HTTP_404_NOT_FOUND)


@override_settings(TOM_FACILITY_CLASSES=['tom_observations.tests.utils.FakeRoboticFacility',
                                         'tom_observations.tests.utils.FakeManualFacility'],
                   TARGET_PERMISSIONS_ONLY=False)
class TestObservationViewsetRowLevelPermissions(APITestCase):
    def setUp(self):
        self.user = User.objects.create(username='testuser')
        self.st = SiderealTargetFactory.create(name='testtarget')

        self.obsr = ObservingRecordFactory.create(target_id=self.st.id, user=self.user, status='PENDING')
        self.obsr2 = ObservingRecordFactory.create(target_id=self.st.id, status='PENDING')
        self.obsr3 = ObservingRecordFactory.create(target_id=self.st.id, status='PENDING')

        assign_perm('tom_observations.view_observationrecord', self.user, self.obsr2)
        assign_perm('tom_observations.change_observationrecord', self.user, self.obsr2)

        self.client.force_login(self.user)

    def test_observation_list_with_permissions(self):
        """Test observation API list endpoint with row-level permissions."""
        response = self.client.get(reverse('api:observations-list'))
        self.assertEqual(response.json()['count'], 2)

    def test_observation_detail_with_permissions(self):
        """Test observation API detail endpoint with row-level permissions."""
        with self.subTest('Test that an ObservationRecord submitted by a user is visible to them.'):
            response = self.client.get(reverse('api:observations-detail', args=(self.obsr.id,)))
            self.assertEqual(response.json()['id'], self.obsr.id)

        with self.subTest('Test that a user can view an ObservationRecord for which they have permissions.'):
            response = self.client.get(reverse('api:observations-detail', args=(self.obsr2.id,)))
            self.assertEqual(response.json()['id'], self.obsr2.id)

        with self.subTest('Test that a user cannot view an ObservationRecord for which they lack permissions.'):
            response = self.client.get(reverse('api:observations-detail', args=(self.obsr3.id,)))
            self.assertContains(response, 'Not found.', status_code=status.HTTP_404_NOT_FOUND)

    def test_observation_submit_with_permissions(self):
        """Test observation API submit endpoint with groups."""
        group = Group.objects.create(name='testgroup')
        group.user_set.add(self.user)
        user2 = User.objects.create(username='testuser2')
        form_data = {
            'target_id': self.st.id,
            'facility': 'FakeRoboticFacility',
            'observation_type': 'OBSERVATION',
            'groups': [{'id': group.id}],
            'observing_parameters': {
                'test_input': 'gnomes'
            }
        }

        response = self.client.post(reverse('api:observations-list'), data=form_data, follow=True)
        self.assertContains(response, 'fakeid', status_code=status.HTTP_201_CREATED)
        self.assertDictContainsSubset({'test_input': 'gnomes'}, response.json()[0]['parameters'])

        # Test that user in testgroup can see observation
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:observations-list'))
        self.assertContains(response, 'fakeid', status_code=status.HTTP_200_OK)

        # Test that user not in testgroup can't see observation
        self.client.force_login(user2)
        response = self.client.get(reverse('api:observations-list'))
        self.assertEqual(response.json()['count'], 0)

    @patch('tom_observations.tests.utils.FakeRoboticFacility.get_observation_status')
    def test_observation_cancel(self, mock_get_status):
        mock_get_status.return_value = {'state': 'CANCELED',
                                        'scheduled_start': timezone.now(),
                                        'scheduled_end': timezone.now()}
        user2 = User.objects.create(username='testuser2')
        self.obsr2.facility = 'FakeRoboticFacility'
        self.obsr2.save()

        response = self.client.patch(reverse('api:observations-cancel', kwargs={'pk': self.obsr2.id}))
        self.assertContains(response, 'CANCELED')

        self.client.force_login(user2)
        response = self.client.patch(reverse('api:observations-cancel', kwargs={'pk': self.obsr2.id}))
        self.assertContains(response, 'Not found.', status_code=status.HTTP_404_NOT_FOUND)



# Generated by Django 3.1.5 on 2021-02-05 18:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_observations', '0011_auto_20200917_0306'),
    ]

    operations = [
        migrations.AlterField(
            model_name='observationrecord',
            name='parameters',
            field=models.JSONField(),
        ),
        migrations.AlterField(
            model_name='observationtemplate',
            name='parameters',
            field=models.JSONField(),
        ),
    ]



from django.db import migrations, models


def copy_tag_to_data_product_type(apps, schema_editor):
    data_products = apps.get_model('tom_dataproducts', 'DataProduct')
    for row in data_products.objects.all():
        row.data_product_type = row.tag
        if row.data_product_type in ['SPECTROSCOPY', 'PHOTOMETRY', 'FITS_FILE', 'IMAGE_FILE']:
            row.data_product_type = row.data_product_type.lower()
        row.save(update_fields=['data_product_type'])


class Migration(migrations.Migration):

    # This migration script does the following:
    # - Adds a column DataProduct.data_product_type
    # - Copies existing tag to data_product_type
    # - Removes column DataProduct.tag

    dependencies = [
        ('tom_dataproducts', '0006_auto_20190912_2013')
    ]

    operations = [
        migrations.AddField(
            model_name='dataproduct',
            name='data_product_type',
            field=models.CharField(blank=False, default='', max_length=50)
        ),

        migrations.RunPython(copy_tag_to_data_product_type, reverse_code=migrations.RunPython.noop),

        migrations.RemoveField(
            model_name='dataproduct',
            name='tag'
        )
    ]



from specutils import Spectrum1D
from astropy.units import Quantity


class SpectrumSerializer():

    def serialize(self, spectrum: Spectrum1D) -> dict:
        """
        Serializes a Spectrum1D in order to store in a ReducedDatum object. The serialization stores only what's
        necessary to rebuild the Spectrum1D--namely, flux and wavelength, and their respective units.

        :param spectrum: Spectrum1D to be serialized
        :type spectrum: specutils.Spectrum1D

        :returns: JSON representation of spectrum
        :rtype: dict
        """
        serialized = {}
        serialized['flux'] = spectrum.flux.value.tolist()
        serialized['flux_units'] = spectrum.flux.unit.to_string()
        serialized['wavelength'] = spectrum.wavelength.value.tolist()
        serialized['wavelength_units'] = spectrum.wavelength.unit.to_string()
        return serialized

    def deserialize(self, spectrum: dict) -> Spectrum1D:
        """
        Constructs a Spectrum1D from the spectrum value stored in a ReducedDatum

        :param spectrum: JSON representation used to construct the Spectrum1D
        :type spectrum: dict

        :returns: Spectrum1D representing the spectrum information
        :rtype: specutil.Spectrum1D
        """
        flux = Quantity(value=spectrum['flux'], unit=spectrum['flux_units'])
        wavelength = Quantity(value=spectrum['wavelength'], unit=spectrum['wavelength_units'])
        spectrum = Spectrum1D(flux=flux, spectral_axis=wavelength)
        return spectrum



from requests.exceptions import HTTPError
from django.core.management.base import BaseCommand
from django.core.exceptions import ObjectDoesNotExist

from tom_alerts import alerts
from tom_targets.models import Target
from tom_dataproducts.models import ReducedDatum


class Command(BaseCommand):
    help = 'Gets and updates time-series data for alert-generated targets from the original alert source.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--target_id',
        )

    def handle(self, *args, **options):
        brokers = alerts.get_service_classes()
        broker_classes = {}
        for broker in brokers:
            broker_classes[broker] = alerts.get_service_class(broker)()

        target = None
        if options['target_id']:
            try:
                targets = [Target.objects.get(pk=options['target_id'])]
            except ObjectDoesNotExist:
                raise Exception('Invalid target id provided')
        else:
            sources = ReducedDatum.objects.filter(source_name__in=broker_classes.keys()).distinct()
            targets = Target.objects.filter(
                id__in=ReducedDatum.objects.filter(
                    source_name__in=sources
                ).values_list('target').distinct())

        failed_records = {}
        for target in targets:
            for class_name, clazz in broker_classes.items():
                try:
                    clazz.process_reduced_data(target)
                except HTTPError:
                    failed_records[class_name] = target.id

        if len(failed_records) == 0:
            return 'Update completed successfully'
        else:
            return 'Update completed with errors: {0}'.format(str(failed_records))
